# -*- coding: utf-8 -*-
'''
Snake class
===========

This `subject` class emulates a single player snake game. 


'''

import curses
from curses import KEY_DOWN, KEY_LEFT, KEY_RIGHT, KEY_UP
from random import randint

from ..valueset import ValueSet
from .mnkboard import MNKBoard
from .mnkgame import MNKGame
from .subject import Subject


def main():
    pass


class Snake(MNKBoard, Subject):
    '''
    Create a single player snake game.

    Attributes
    ----------
        is_terminated: whether the game finished or not.
        possible_actions: a list of possible actions.

    Methods
    -------
        take_effect: set a piece of the specified player on the specified square of the board.
        reset: clear the board.
    '''
    def __init__(self, **kwargs):
        '''
        Initialize an instance of snake game.

        Arguments
        ---------
            m: number of rows (default=10)
            n: number of columns (default=10)
        '''
        self._m = 10
        self._n = 10
        self.set_params(**kwargs)
        super().__init__(**kwargs)
        self.reset()

    @property
    def is_terminated(self):
        '''Return True if the game is finished.'''
        return (self._snake[0] in self._snake[1:])
    
    @property
    def possible_actions(self):
        '''Return the moves as ValueSet (left, none, right).'''
        return ValueSet(['left', 'none', 'right'])

    def take_effect(self, action, _id=None):
        '''
        Move the snake on the board.

        Arguments
        ---------
            _id: ID of the player. (Not used in the code)
            action: one of three possible actions: left, none, right.
        ''' 
        self._win.border(0)
        self._win.addstr(0, 2, 'Score : ' + str(self._score) + ' ')                # Printing 'Score' and
        self._win.addstr(0, 27, ' SNAKE ')                                   # 'SNAKE' strings
        self._win.timeout(150 - int(len(self._snake)/5 + len(self._snake)/10)%120)          # Increases the speed of Snake as its length increases
        
        # prevKey = self._key                                                  # Previous key pressed
        _ = self._win.getch()
        if action == 'left':
            if self._key == KEY_LEFT:
                self._key = KEY_DOWN
            elif self._key == KEY_DOWN:
                self._key = KEY_RIGHT
            elif self._key == KEY_RIGHT:
                self._key = KEY_UP
            elif self._key == KEY_UP:
                self._key = KEY_LEFT
        elif action == 'right':
            if self._key == KEY_LEFT:
                self._key = KEY_UP
            elif self._key == KEY_UP:
                self._key = KEY_RIGHT
            elif self._key == KEY_RIGHT:
                self._key = KEY_DOWN
            elif self._key == KEY_DOWN:
                self._key = KEY_LEFT

        # if key == ord(' '):                                            # If SPACE BAR is pressed, wait for another
        #     key = -1                                                   # one (Pause/Resume)
        #     while key != ord(' '):
        #         key = win.getch()
        #     key = prevKey
        #     continue

        # if key not in [KEY_LEFT, KEY_RIGHT, KEY_UP, KEY_DOWN, 27]:     # If an invalid key is pressed
        #     key = prevKey

        # Calculates the new coordinates of the head of the snake. NOTE: len(snake) increases.
        # This is taken care of later at [1].
        self._snake.insert(0, [self._snake[0][0] + (self._key == KEY_DOWN and 1) + (self._key == KEY_UP and -1), self._snake[0][1] + (self._key == KEY_LEFT and -1) + (self._key == KEY_RIGHT and 1)])
        if self._snake[0][0] == 0: self._snake[0][0] = self._m - 2
        if self._snake[0][1] == 0: self._snake[0][1] = self._n - 2
        if self._snake[0][0] == self._m: self._snake[0][0] = 1
        if self._snake[0][1] == self._n: self._snake[0][1] = 1
        # print(self._snake[0][0], self._snake[0][1], (self._snake[0][0] * 6 + self._snake[0][1]))
        self.set_piece(1, row=self._snake[0][0], column=self._snake[0][1])
        if self._snake[0] == self._food:                                            # When snake eats the food
            self._food = []
            self._score += 1
            while self._food == []:
                self._food = [randint(1, self._m-2), randint(1, self._n-2)]                 # Calculating next food's coordinates
                if self._food in self._snake: self._food = []
            self._win.addch(self._food[0], self._food[1], '*')
            self.set_piece(2, row=self._food[0], column=self._food[1])
        else:
            last = self._snake.pop()                                          # [1] If it does not eat the food, length decreases
            self._win.addch(last[0], last[1], ' ')
            self.clear_square(row=last[0], column=last[1])
        self._win.addch(self._snake[0][0], self._snake[0][1], '#')
        return self._score

    def reset(self):
        '''Clear the board, reset the snake and the fruit and update board_status.'''
        MNKBoard.reset(self)
        curses.initscr()
        self._win = curses.newwin(20, 60, 0, 0)
        self._win.keypad(1)
        curses.noecho()
        curses.curs_set(0)
        self._win.border(0)
        self._win.nodelay(1)
        self._key = KEY_RIGHT                                                    # Initializing values
        self._score = 0
        self._snake = [[self._m // 2, self._n // 2 + 1],
                       [self._m // 2, self._n // 2],
                       [self._m // 2, self._n // 2 - 1]]                                     # Initial snake co-ordinates
        self._food = [self._m // 2 + 1, self._n // 2]                                                     # First food co-ordinates
        self._win.addch(self._food[0], self._food[1], '*')                       # Prints the food
        super().__init__(self)
        for location in self._snake:
            self.set_piece(1, row=location[0], column=location[1])
        self.set_piece(2, row=self._food[0], column=self._food[1])

    def __repr__(self):
        return 'Snake'

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
'''
environments module for reinforcement learning
==============================================

This module contains classes that provides tools to load/save objects and
environments, add/remove `agents`/`subjects`, assign `agents` to `subjects` and
run models.

Classes
-------
Environment:
    The base class of all environment classes.

EnvironmentStaticMap:
    An environment with static interaction sequence.

Task:
    A class to define a `ReiL` task.

Session:
    A class that accepts `Agent`s, `Subject`s, and `Demon`s, and
    runs a set of `Task`s.

SessionBuilder:
    A tool to create a `Session` from one or a list of `YAML` config files.
'''

from .environment import (  # noqa: W0611
    Environment, EntityType, EntityGenType, AgentSubjectTuple)
from .environment_static_map import EnvironmentStaticMap  # noqa: W0611
from .task import Task  # noqa: W0611
from .session import Session  # noqa: W0611
from .session_builder import SessionBuilder  # noqa: W0611



# -*- coding: utf-8 -*-
'''
EndlessBuffer class
===================

A `Buffer` without size limit.
'''

from typing import Dict, List, Optional, Union

from reil.datatypes.buffers.buffer import Buffer, PickModes, T1, T2


class EndlessBuffer(Buffer[T1, T2]):
    '''
    A `Buffer` without size limit.

    Extends `Buffer` class.
    '''

    def __init__(
            self,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None) -> None:
        '''
        Arguments
        ---------
        buffer_names:
            A list containing the names of buffer queues.

        pick_mode:
            The default mode to pick items from the list.
        '''
        self.setup(buffer_names=buffer_names, pick_mode=pick_mode)

    def setup(
            self, buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None,
            clear_buffer: Optional[bool] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer. This argument is only available for
            signature consistency. Assigning it has no effect.

        buffer_names:
            A list containing the names of buffer elements.

        pick_mode:
            The default mode to pick items from the list.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.

        Raises
        ------
        ValueError:
            Cannot modify `buffer_size`. The value is already set.

        ValueError:
            Cannot modify `buffer_names`. The value is already set.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        super().setup(
            buffer_names=buffer_names, pick_mode=pick_mode,
            clear_buffer=clear_buffer)

    def add(self, data: Dict[str, Union[T1, T2]]) -> None:
        '''
        Add a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.
        '''
        if self._buffer is not None:
            self._buffer_index += 1
            for key, v in data.items():
                self._buffer[key].append(v)  # type: ignore
            self._count += 1

        else:
            raise RuntimeError('Buffer is not set up.')

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        super().reset()
        if self._buffer_names is not None:
            self._buffer = {name: []  # type: ignore
                            for name in self._buffer_names}
        else:
            self._buffer = None  # type: ignore



import logging
import pathlib
from math import ceil, log10
from typing import Dict, Literal, Optional, Tuple, Union

from reil.datatypes.interaction_protocol import InteractionProtocol
from reil.environments.environment_static_map import EnvironmentStaticMap
from reil.utils.output_writer import OutputWriter

# TODO: Documentation
# TODO: what should we do with `save` and `iterations`?!


class Task:
    def __init__(
            self, name: str, path: Union[pathlib.PurePath, str],
            agent_training_triggers: Dict[str, Literal[
                'none', 'termination',
                'state', 'action', 'reward']],
            interaction_sequence: Tuple[InteractionProtocol, ...],
            start_iteration: int = 0, max_iterations: int = 1,
            writer: Optional[OutputWriter] = None,
            save_iterations: bool = True):
        self._name = name
        self._path = path
        self.max_iterations = max_iterations
        self._start_iteration = start_iteration
        self._agent_training_triggers = agent_training_triggers
        self._writer = writer

        # self._subjects = subjects

        self._save_iterations = save_iterations
        if save_iterations:
            self._filename_format = f'{{}}_{{:0{ceil(log10(max_iterations))}}}'
        else:
            self._filename_format = f'{{}}'

        for protocol in interaction_sequence:
            if protocol.unit != 'iteration':
                logging.warning(
                    'Interaction protocol unit should be "iteration". '
                    'Current unit might have unintended consequences.')

        self._interaction_sequence = interaction_sequence

    def run_file(
        self, environment_filename: str,
        path: pathlib.PurePath, iteration: int
    ) -> EnvironmentStaticMap:
        return self.run_env(EnvironmentStaticMap.from_pickle(
            environment_filename, path), iteration)

    def run_env(
            self, env: EnvironmentStaticMap,
            iteration: int) -> EnvironmentStaticMap:
        env.interaction_sequence = self._interaction_sequence

        for agent, trigger in self._agent_training_triggers.items():
            env._agents[agent]._training_trigger = trigger

        for protocol in self._interaction_sequence:
            env._iterations[protocol.subject.name] = iteration

        env.simulate_pass()
        if self._writer:
            rep = env.report_statistics(unstack=True, reset_history=True)
            self._writer.write_stats_output(rep)

        return env

        # self.simulate(env, self._writer)
        # f, p = env.save(
        #     filename=self._filename_format.format(self._name, iteration + 1),



# -*- coding: utf-8 -*-
'''
CircularBuffer class
====================

A `Buffer` that overflows!
'''

from typing import Dict, List, Optional, Tuple, Union

from reil.datatypes.buffers.buffer import Buffer, PickModes, T1, T2


class CircularBuffer(Buffer[T1, T2]):
    '''
    A `Buffer` that overflows.

    Extends `Buffer` class.
    '''

    def __init__(
            self, buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None) -> None:
        self._buffer_full: bool = False
        super().__init__(buffer_size=buffer_size, buffer_names=buffer_names,
                         pick_mode=pick_mode)

    def add(self, data: Dict[str, Union[T1, T2]]) -> None:
        '''
        Add a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.

        Notes
        -----
        If the buffer is full, new items will be writen over the oldest one.
        '''
        try:
            super().add(data)
        except IndexError:
            self._buffer_full = True
            self._buffer_index = -1
            super().add(data)

        # the size does not change if buffer is full.
        self._count -= self._buffer_full

    def _pick_old(
        self, count: int
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return the oldest items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.
        '''
        if self._buffer_full:
            self._buffer_size: int
            slice_pre = slice(self._buffer_index + 1,
                              self._buffer_index + count + 1)
            slice_post = slice(
                max(0, count - (self._buffer_size - self._buffer_index) + 1))

            return {name: tuple(
                buffer[slice_pre] + buffer[slice_post])  # type: ignore
                for name, buffer in self._buffer.items()}
        else:
            return super()._pick_old(count)

    def _pick_recent(
        self, count: int
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return the most recent items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.
        '''
        if count - self._buffer_index <= 1 or not self._buffer_full:
            return super()._pick_recent(count)
        elif self._buffer is not None:
            slice_pre = slice(-(count - self._buffer_index - 1), None)
            slice_post = slice(self._buffer_index + 1)
            return {name: tuple(
                buffer[slice_pre] + buffer[slice_post])  # type: ignore
                for name, buffer in self._buffer.items()}
        else:
            raise RuntimeError('Buffer is not set up.')

    def _pick_all(self) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return all items in the buffer.
        '''
        if self._buffer_full:
            self._buffer: Dict[str, Union[List[T1], List[T2]]]
            slice_pre = slice(self._buffer_index + 1, None)
            slice_post = slice(self._buffer_index + 1)
            return {name: tuple(
                buffer[slice_pre] + buffer[slice_post])  # type: ignore
                for name, buffer in self._buffer.items()}
        else:
            return super()._pick_all()

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        super().reset()
        self._buffer_full = False



from .aurora import Aurora  # noqa: W0611
from .intermountain import Intermountain  # noqa: W0611
from .iwpc import IWPC  # noqa: W0611
from .lenzini import Lenzini  # noqa: W0611
from .ravvaz_dosing_protocols import AAA, CAA, PGAA, PGPGA, PGPGI  # noqa: W0611



# -*- coding: utf-8 -*-
'''
agents module for healthcare
============================

This module provides agents in healthcare context.

Classes
-------
WarfarinAgent
    an agent based on Ravvaz et al (2016) paper for Warfarin Dosing

'''
from .warfarin_agent import WarfarinAgent  # noqa: W0611



# -*- coding: utf-8 -*-
'''
Risk class
==========

This `subject` class emulates dice throwing in Risk board game. 


'''

from random import randrange

from reil.valueset import ValueSet
from reil.subjects.subject import Subject


def main():
    board = Risk(pieces=[5, 5])
    player = {}
    player['P1'] = board.register('P1')
    player['P2'] = board.register('P2')
    while not board.is_terminated:
        board.take_effect(player['P1'], ValueSet(2))
        board.take_effect(player['P2'], ValueSet(1))
        print(board._pieces)


class Risk(Subject):
    '''

    Attributes
    ----------
        is_terminated: whether the game finished or not.
        possible_actions: a list of possible actions (how many pieces to play).

    Methods
    -------
        register: register a new player and return its ID or return ID of an existing player.
        take_effect: roll the dice and return the result.
        ??? set_piece: set a piece of the specified player on the specified square of the board.
        reset: clear the conditions.
    '''
    # _board is a row vector. (row, column) and index start from 0
    # _board_status: None: no winner yet,
    #                1..players: winner,
    #                0: stall,
    #               -1: illegal board
    def __init__(self, **kwargs):
        '''
        Initialize an instance of risk.

        Arguments
        ---------
            pieces: an array containing the number of peices for players 1 and 2 (default=[3, 2])
        '''
        self._pieces = [3, 2]
        self._turn = 0
        self.set_params(**kwargs)
        super().__init__(**kwargs)

        self._temp = 0

        self._initial_pieces = list(self._pieces)

    @property
    def is_terminated(self):
        '''Return True if no moves is left.'''
        return (self._pieces[0] <= 1) | (self._pieces[1] <= 0)

    @property
    def possible_actions(self):
        '''Return a list of indexes of empty squares.'''
        max_move = min(self._pieces[self._turn], [3, 2][self._turn])
        if self._turn == 1:
            max_move = min(self._temp, max_move)
        try:
            return ValueSet(list(range(1, max_move + 1)), min=1, max=[3, 2][self._turn]).as_valueset_array()
        except TypeError:
            return ValueSet(0, min=0, max=[3, 2][self._turn]).as_valueset_array()

    @property
    def state(self):
        return ValueSet(self._pieces, min=[0, 0], max=self._initial_pieces)

    def register(self, player_name):
        '''
        Register an agent and return its ID.
        
        If the agent is new, a new ID is generated and the agent_name is added to agent_list.
        Arguments
        ---------
            agent_name: the name of the agent to be registered.
        '''
        if len(self._agent_list)<2:
            return Subject.register(self, player_name)

    def take_effect(self, action, _id=None):
        '''
        Set a piece for the given player on the board.

        Arguments
        ---------
            _id: ID of the player who sets the piece.
            action: the location in which the piece is set. Can be either in index format or row column format.
        '''
        if self._turn == 0:
            self._temp = action.value[0]
            self._turn = 1
            return 0
        else:
            # reward = 0
            p1 = list(randrange(6) for _ in range(self._temp))
            p2 = list(randrange(6) for _ in range(action.value[0]))
            for _ in range(min(self._temp, action.value[0])):
                if max(p1)<=max(p2):
                    self._pieces[0] -= 1
                    # reward += 1
                else:
                    self._pieces[1] -= 1
                    # reward -= 1
                p1.remove(max(p1))
                p2.remove(max(p2))

            self._turn = 0
            # return reward
            if self._pieces[0] <= 0:  # player 1 wins, so player 2 loses
                return 1
            elif self._pieces[1] <= 0:
                return -1
            else:
                return 0

    def reset(self):
        '''Clear the board and update board_status.'''
        self._pieces = list(self._initial_pieces)
        self._turn = 0
        self._temp = 0

    def __repr__(self):
        return 'Risk'

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
'''
Learner class
=============

The base class for all `learner` classes.
'''
from typing import Any, Generic, Tuple, TypeVar

from reil import reilbase
from reil.datatypes.feature import FeatureArray
from reil.learners.learning_rate_schedulers import (ConstantLearningRate,
                                                    LearningRateScheduler)

LabelType = TypeVar('LabelType')


class Learner(reilbase.ReilBase, Generic[LabelType]):
    '''
    The base class for all `learner` classes.
    '''
    def __init__(self,
                 learning_rate: LearningRateScheduler,
                 **kwargs: Any) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the learning rate
            based on iteration. If any scheduler other than constant is
            provided, the model uses the `new_rate` method of the scheduler to
            determine the learning rate at each iteration.
        '''
        super().__init__(**kwargs)
        self._learning_rate = learning_rate

    @classmethod
    def _empty_instance(cls):
        return cls(learning_rate=ConstantLearningRate(0.0))

    def predict(self, X: Tuple[FeatureArray, ...]) -> Tuple[LabelType, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        raise NotImplementedError

    def learn(
            self, X: Tuple[FeatureArray, ...], Y: Tuple[LabelType, ...]
    ) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        raise NotImplementedError



# -*- coding: utf-8 -*-
'''
PatientWarfarinRavvaz class
===========================

A warfarin patient class with features and parameters of Ravvaz et al. 2016.

Features included in this model are:
* age
* weight
* height
* gender
* race
* tobaco
* amiodarone
* fluvastatin
* CYP2C9
* VKORC1
* MTT_1
* MTT_2
* cyp_1_1
* V1
* V2
* EC_50
'''
import math
from typing import Any, Literal

from reil.datatypes.feature import FeatureGenerator
from reil.healthcare.patient import Patient
from reil.healthcare.mathematical_models import HealthMathModel
from reil.healthcare.mathematical_models.hamberg_pkpd import HambergPKPD
from reil.utils.functions import (random_categorical,
                                  random_lognormal_truncated,
                                  random_normal_truncated)

# pre-computing to gain some speed boost!
log_EC_50_GG = math.log(HambergPKPD._EC_50_GG)
log_EC_50_GA = math.log(HambergPKPD._EC_50_GA)
log_EC_50_AA = math.log(HambergPKPD._EC_50_AA)
sqrt_omega_EC_50 = math.sqrt(HambergPKPD._omega_EC_50)


class PatientWarfarinRavvaz(Patient):
    def __init__(self,
                 model: HealthMathModel,
                 randomized: bool = True,
                 allow_missing_genotypes: bool = True,
                 **feature_values: Any) -> None:
        '''
        Parameters
        ----------
        model:
            A `HealthMathModel` to be used to model patient's behavior.

        randomized:
            Whether patient characteristics and model parameters should be
            generated randomly or deterministically.

        feature_values:
            Keyword arguments by which some of the `features` of the patient
            can be determined. For example, if "age" is one of the features,
            age=40.0 will set the initial age to 40.0.
        '''
        self.feature_gen_set = {
            'age': FeatureGenerator[float].numerical(
                name='age',  # (years) Aurora population
                lower=18.0, upper=150.0, mean=67.30, stdev=13.43,
                generator=random_normal_truncated,
                randomized=randomized),
            'weight': FeatureGenerator[float].numerical(
                name='weight',  # (lb) Aurora population
                lower=70.0, upper=500.0, mean=199.24, stdev=54.71,
                generator=random_normal_truncated,
                randomized=randomized),
            'height': FeatureGenerator[float].numerical(
                name='height',  # (in) Aurora population
                lower=45.0, upper=85.0, mean=66.78, stdev=4.31,
                generator=random_normal_truncated,
                randomized=randomized),
            'gender': FeatureGenerator[Literal['Female', 'Male']].categorical(
                name='gender',  # Aurora population
                categories=('Female', 'Male'),
                probabilities=(0.5314, 0.4686),
                generator=random_categorical,
                randomized=randomized),
            'race': FeatureGenerator[Literal[
                'White', 'Black', 'Asian',
                'American Indian', 'Pacific Islander'
            ]].categorical(
                name='race',  # Aurora Avatar Population
                categories=('White', 'Black', 'Asian',
                            'American Indian', 'Pacific Islander'),
                probabilities=(0.9522, 0.0419, 0.0040, 0.0018, 1e-4),
                generator=random_categorical,
                randomized=randomized),
            'tobaco': FeatureGenerator[Literal['No', 'Yes']].categorical(
                name='tobaco',  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.9067, 0.0933),
                generator=random_categorical,
                randomized=randomized),
            'amiodarone': FeatureGenerator[Literal['No', 'Yes']].categorical(
                name='amiodarone',  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.8849, 0.1151),
                generator=random_categorical,
                randomized=randomized),
            'fluvastatin': FeatureGenerator[Literal['No', 'Yes']].categorical(
                name='fluvastatin',  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.9998, 0.0002),
                generator=random_categorical,
                randomized=randomized),
            'CYP2C9': FeatureGenerator[Literal[
                '*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3', '*3/*3'
            ]].categorical(
                name='CYP2C9',  # Aurora Avatar Population
                categories=('*1/*1', '*1/*2', '*1/*3',
                            '*2/*2', '*2/*3', '*3/*3'),
                probabilities=(0.6739, 0.1486, 0.0925, 0.0651, 0.0197, 2e-4),
                generator=random_categorical,
                randomized=randomized,
                allow_missing=allow_missing_genotypes),
            'VKORC1': FeatureGenerator[Literal[
                'G/G', 'G/A', 'A/A'
            ]].categorical(
                name='VKORC1',  # Aurora Avatar Population
                categories=('G/G', 'G/A', 'A/A'),
                probabilities=(0.3837, 0.4418, 0.1745),
                generator=random_categorical,
                randomized=randomized,
                allow_missing=allow_missing_genotypes),

            'MTT_1': FeatureGenerator[float].numerical(
                name='MTT_1',  # (hours) Hamberg PK/PD
                mean=math.log(HambergPKPD._MTT_1),
                stdev=math.sqrt(HambergPKPD._omega_MTT_1),
                generator=random_lognormal_truncated,
                randomized=randomized),
            'MTT_2': FeatureGenerator[float].numerical(
                name='MTT_2',  # (hours) Hamberg PK/PD
                # Hamberg et al. (2007) - Table 4
                mean=math.log(HambergPKPD._MTT_2),
                stdev=math.sqrt(HambergPKPD._omega_MTT_2),
                generator=random_lognormal_truncated,
                randomized=randomized),
            'CL_S_cyp_1_1': FeatureGenerator[float].numerical(
                name='CL_S_cyp_1_1',  # (l/h) Hamberg PK/PD
                mean=math.log(HambergPKPD._CL_s_1_1),
                stdev=math.sqrt(HambergPKPD._omega_CL_s),
                generator=random_lognormal_truncated,
                randomized=randomized),
            'V1': FeatureGenerator[float].numerical(
                name='V1',  # (L) Volume in central compartment
                mean=math.log(HambergPKPD._V1),
                stdev=math.sqrt(HambergPKPD._omega_V1),
                generator=random_lognormal_truncated,
                randomized=randomized),
            'V2': FeatureGenerator[float].numerical(
                name='V2',  # (L) volume in peripheral compartment
                mean=math.log(HambergPKPD._V2),
                stdev=math.sqrt(HambergPKPD._omega_V2),
                generator=random_lognormal_truncated,
                randomized=randomized),

            # 'EC_50': FeatureGenerator.numerical(
            #     name='EC_50',  # (mg/L) Hamberg PK/PD
            #     stdev=math.sqrt(0.409),
            #     generator=random_lognormal_truncated,
            #     randomized=randomized),

            # 'sensitivity': FeatureGenerator.categorical(
            #     name='sensitivity',
            #     categories=('normal', 'sensitive', 'highly sensitive'),
            #     generator=lambda f: f.value,
            #     randomized=randomized)
        }

        self._sensitivity_gen = FeatureGenerator[Literal[
            'normal', 'sensitive', 'highly sensitive'
        ]].categorical(
            name='sensitivity',
            categories=('normal', 'sensitive', 'highly sensitive'))

        # Since EC_50 is not set (it depends on other features),
        # super().__init__() fails to setup the model.
        # I catch it, generate EC_50 and set up the model.
        self._randomized = randomized
        try:
            super().__init__(model, **feature_values)
        except KeyError:
            self._generate_EC_50()
            self._generate_sensitivity()
            self._model.setup(**self.feature_set)

    def generate(self) -> None:
        self.feature_gen_set.pop('EC_50')
        super().generate()

        self._generate_EC_50()
        self._generate_sensitivity()
        self._model.setup(**self.feature_set)

    def _generate_EC_50(self) -> None:
        vkorc1 = self.feature_set['VKORC1'].value
        if vkorc1 == 'G/G':
            mean = log_EC_50_GG
        elif vkorc1 in ('G/A', 'A/G'):
            mean = log_EC_50_GA
        else:  # 'A/A'
            mean = log_EC_50_AA

        self.feature_gen_set['EC_50'] = FeatureGenerator[float].numerical(
            name='EC_50',  # (mg/L) Hamberg PK/PD
            mean=mean,
            stdev=sqrt_omega_EC_50,
            generator=random_lognormal_truncated,
            randomized=self._randomized)

        self.feature_set['EC_50'] = self.feature_gen_set['EC_50']()

    def _generate_sensitivity(self):
        combo = (str(self.feature_set['CYP2C9'].value) +
                 str(self.feature_set['VKORC1'].value))

        if combo in ('*1/*1G/G', '*1/*2G/G', '*1/*1G/A'):
            s = 'normal'
        elif combo in ('*1/*2G/A', '*1/*3G/A', '*2/*2G/A',
                       '*2/*3G/G', '*1/*3G/G', '*2/*2G/G',
                       '*1/*2A/A', '*1/*1A/A'):
            s = 'sensitive'
        elif combo in ('*3/*3G/G',
                       '*3/*3G/A', '*2/*3G/A',
                       '*3/*3A/A', '*2/*3A/A', '*2/*2A/A', '*1/*3A/A'):
            s = 'highly sensitive'
        else:
            raise ValueError(
                f'Unknown CYP2C9 and VKORC1 combination: {combo}.')

        self.feature_set['sensitivity'] = self._sensitivity_gen(s)



# -*- coding: utf-8 -*-
'''
WarfarinAgent class
===================

An agent for warfarin modeling based on the protocols defined
in Ravvaz et al (2017).
'''

from typing import Any, Literal, Optional, Tuple

from reil import agents
from reil.datatypes.feature import FeatureArray, FeatureGenerator
from reil.healthcare.dosing_protocols.warfarin import (AAA, CAA, PGAA, PGPGA,
                                                       PGPGI)


class WarfarinAgent(agents.NoLearnAgent):
    '''
    An `agent` that prescribes dose for a warfarin `subject`,
    based on the dosing protocols defined in Ravvaz et al (2017).
    '''

    def __init__(self,
                 study_arm: Literal['aaa', 'caa', 'pgaa',
                                    'pgpgi', 'pgpga'] = 'aaa',
                 dose_range: Tuple[float, float] = (0.0, 15.0),
                 interval_range: Tuple[int, int] = (1, 28),
                 **kwargs: Any):
        '''
        Arguments
        ---------
        study_arm:
            One of available study arms: aaa, caa, pgaa, pgpgi, pgpga
        '''
        super().__init__(**kwargs)

        if study_arm.lower() in ['aaa', 'ravvaz aaa', 'ravvaz_aaa']:
            self._protocol = AAA()
        elif study_arm.lower() in ['caa', 'ravvaz caa', 'ravvaz_caa']:
            self._protocol = CAA()
        elif study_arm.lower() in ['pgaa', 'ravvaz pgaa', 'ravvaz_pgaa']:
            self._protocol = PGAA()
        elif study_arm.lower() in ['pgpgi', 'ravvaz pgpgi', 'ravvaz_pgpgi']:
            self._protocol = PGPGI()
        elif study_arm.lower() in ['pgpga', 'ravvaz pgpga', 'ravvaz_pgpga']:
            self._protocol = PGPGA()

        self._dose_gen = FeatureGenerator[float].numerical(
            name='dose', lower=dose_range[0], upper=dose_range[1])
        self._interval_gen = FeatureGenerator[float].numerical(
            name='interval', lower=interval_range[0], upper=interval_range[1])

    def act(self,
            state: FeatureArray,
            subject_id: int,
            actions: Optional[Tuple[FeatureArray, ...]] = None,
            iteration: Optional[int] = 0) -> FeatureArray:
        '''
        Generate the dosing `action` based on the `state` and current dosing
        protocol.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        iteration:
            The iteration in which the agent is acting.

        Returns
        -------
        :
            The action
        '''
        patient = state.value
        patient['day'] += 1

        decision = self._protocol.prescribe(patient)
        dose = decision.dose
        interval = decision.duration

        if interval is None:
            raise ValueError(f'None duration received from {self._protocol}.')

        return FeatureArray([
            self._dose_gen(min(dose, self._dose_gen.upper or dose)),
            self._interval_gen(
                min(interval, self._interval_gen.upper or interval))
        ])

    def reset(self):
        '''Resets the agent at the end of a learning iteration.'''
        self._protocol.reset()

    def __repr__(self) -> str:
        try:
            return super().__repr__() + f' arm: {self._protocol}'
        except NameError:
            return super().__repr__()

# Implementation that matches RAVVAZ dataset.
# -------------------------------------------
#
# def aurora(self, patient: Dict[str, Any]) -> float:
#     if self._red_flag:
#         if self._retest_day > patient['day']:
#             return_value = 0.0
#         elif patient['INR_history'][-1] > 3.0:
#             self._retest_day = patient['day'] + 2
#             return_value = 0.0
#         else:
#             self._red_flag = False
#             self._retest_day = patient['day'] + 7
#             return_value = self._dose
#         return return_value

#     next_test = 2
#     if patient['day'] <= 2:
#         self._dose = 10.0 if patient['age'] < 65.0 else 5.0
#     elif patient['day'] <= 4:
#         day_2_INR = (patient['INR_history'][-1] if patient['day'] == 3
#                      else patient['INR_history'][-2])
#         if day_2_INR >= 2.0:
#             self._dose = 5.0
#             if day_2_INR <= 3.0:
#                 self._early_therapeutic = True

#             self._number_of_stable_days, next_test = \
#                 self._aurora_retesting_table(patient['INR_history'][-1],
#                                              self._number_of_stable_days,
#                                              self._early_therapeutic)
#         else:
#             self._dose, next_test, _, _ = self._aurora_dosing_table(
#                 day_2_INR, self._dose)
#     else:
#         self._number_of_stable_days, next_test = \
#             self._aurora_retesting_table(
#               patient['INR_history'][-1],
#               self._number_of_stable_days,
#               self._early_therapeutic)

#     if next_test == -1:
# #         self._early_therapeutic = False
#         self._number_of_stable_days = 0
#         self._dose, next_test, self._skip_dose, self._red_flag = \
#           self._aurora_dosing_table(patient['INR_history'][-1], self._dose)

#     self._retest_day = patient['day'] + next_test

#     return self._dose if self._skip_dose == 0 else 0.0

# def _aurora_dosing_table(self,
#     current_INR: float, dose: float) -> Tuple[float, int, int, bool]:
#     skip_dose = 0
#     red_flag = False
#     if current_INR < 1.50:
#         dose = dose * 1.15
#         next_test = 7
#     elif current_INR < 1.80:
#         dose = dose * 1.10
#         next_test = 7
#     elif current_INR < 2.00:
#         dose = dose * 1.075
#         next_test = 7
#     elif current_INR <= 3.00:
#         next_test = 28
#     elif current_INR < 3.40:
#         dose = dose * 0.925
#         next_test = 7
#     elif current_INR < 4.00:
#         dose = dose * 0.9
#         next_test = 7
#     elif current_INR <= 5.00:
#         skip_dose = 1  # 2
#         dose = dose * 0.875
#         next_test = 7
#     else:
#         red_flag = True
#         next_test = 2
#         dose = dose * 0.85

#     return dose, next_test, skip_dose, red_flag

# def _aurora_retesting_table(self,
#   current_INR: float, number_of_stable_days: int,
#   early_therapeutic: bool = False) -> Tuple[int, int]:
#     # next_test = {0: 1, 1: 1, 2: 5, 7: 7, 14: 14, 28: 28}
#     # NOTE: When patient gets into the range early on
#     #      (before the maintenance period), we have 1: 2,
#     # but when patient is in maintenance,\
#     # we have 1: 6. I track the former in the main aurora method.

#     if early_therapeutic:
#         next_test = {0: 1, 1: 2, 3: 4, 7: 6, 13: 6, 19: 13, 32: 27}
#         max_gap = 32
#     else:
#         next_test = {0: 1, 1: 6, 7: 6, 13: 13, 26: 27}
#         max_gap = 26
#     if 2.0 <= current_INR <= 3.0:
#         number_of_stable_days = min(
#             number_of_stable_days + next_test[number_of_stable_days],
#             max_gap)
#     else:
#         return -1, -1

#     return number_of_stable_days, next_test[number_of_stable_days]



import re
import unittest
from pathlib import Path

import pandas as pd
from reil.agents import WarfarinAgent
from reil.subjects.healthcare import Warfarin


class testWarfarinAgent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._dataset_path = (
            'C:/Users/Sadjad/OneDrive - University of Iowa/'
            '2. Current Research/Anticoagulation/dateset')
        cls._patient_profiles_filename = (
            './UIOWA_KR_clinicalavatars10k_ageUB.csv')
        cls._patient_profiles = pd.read_csv(
            Path(cls._dataset_path, cls._patient_profiles_filename),
            index_col='ID')
        cls._files_in_arm_end = 2  # 41
        cls._study_arms = {
            'aaa': './arm1/ahc_avatars_ahc_algorithm',
            'caa': './arm2/ahc_avatars_clinitialIWPC_ahc_ahc_algorithm',
            'pgaa': './arm3/ahc_avatars_pginitialIWPC_ahc_ahc_algorithm',
            'pgpgi': (
                './arm4/ahc_avatars_eupact'
                'Initial_eupactAlteration_Intermt_algorithm'),
            'pgpga': (
                './arm5/ahc_avatars_eupact'
                'Initial_eupactAlteration_ahc_algorithm')
        }

        cls._rounding_precision = {
            'aaa': 9,
            'caa': 5,
            'pgaa': 2,
            'pgpgi': 9,
            'pgpga': 9
        }

        cls._exceptions_IDs = {
            'aaa': [1000038, 1000055, 1000086, 1000111, 1000143, 1000149,
                    1000171, 1000207, 1000211, 1000219],
            'caa': [1000149, 1000190],
            'pgaa': [],
            'pgpgi': [],
            'pgpga': []
        }

    def test_aaa(self):
        self._run_test('aaa')

    def test_caa(self):
        self._run_test('caa')

    def test_pgaa(self):
        self._run_test('pgaa')

    def test_pgpgi(self):
        self._run_test('pgpgi')

    def test_pgpga(self):
        self._run_test('pgpga')

    def _run_test(self, arm: str) -> None:
        output_files = [f'{self._study_arms[arm]}_{i}.txt' for i in range(
            1, self._files_in_arm_end)]
        raised = False

        for filename in output_files:
            df = pd.read_csv(Path(self._dataset_path, filename), delimiter='|')
            counter_indexes = list(int(re.findall(
                r'\d+$', col)[0])
                for col in df.filter(regex=r'ID\.(\d+)$').columns)
            counter_start = min(counter_indexes)
            counter_end = max(counter_indexes) + 1
            ID_list = dict((counter, df.filter(
                regex=f'ID\.{counter}$').iat[0, 0])
                for counter in range(counter_start, counter_end))

            for counter in range(counter_start, counter_end):
                # , ['AGE', 'CYP2C9', 'VKORC1']]
                patient_info = self._patient_profiles.loc[ID_list[counter]]
                if patient_info.name in self._exceptions_IDs[arm]:
                    continue

                dose_info = df.filter(regex=f'(INR|Dose|Check)\.{counter}$').rename(
                    columns=lambda x: re.findall(r'^\w+', x)[0])

                w = Warfarin(
                    characteristics={
                        'age': patient_info.AGE, 'weight': patient_info.WEIGHT,
                        'height': patient_info.HEIGHT,
                        'CYP2C9': patient_info.CYP2C9, 'VKORC1': patient_info.VKORC1,
                        'gender': patient_info.GENDER,
                        'race': {
                            'White': 'White',
                            'Black or African American': 'Black',
                            'American Indian or Alaskan Native': 'American Indian',
                            'Asian': 'Asian',
                            'Native Hawaiian/Other Pacific Islander': 'Pacific Islander'
                        }[patient_info.RACE],
                        'tobaco': {'N': 'No', 'Y': 'Yes'}[patient_info.SMOKER],
                        'amiodarone': {'N': 'No', 'Y': 'Yes'}[patient_info.AMI],
                        'fluvastatin': {'N': 'No', 'Y': 'Yes'}[patient_info.FLUVASTATIN]},
                    patient_selection='fixed',
                    ex_protocol_current={
                        'state': 'extended', 'possible_actions': 'standard', 'take_effect': 'no_reward'},
                             dose_history_length=5)
                a = WarfarinAgent(study_arm=arm)
                w._decision_points_INR_history[0] = 1.0
                for i in range(dose_info.shape[0]):
                    action = a.act(w.state)
                    w._decision_points_dose_history[w._decision_points_index] = dose_info['Dose'][i]
                    w._decision_points_index += 1
                    w._decision_points_INR_history[w._decision_points_index] = dose_info['INR'][i]
                    w._day += 1

                    try:
                        self.assertEqual(round(action[0].value, self._rounding_precision[arm]), round(dose_info['Dose'][i], self._rounding_precision[arm]),
                                         msg=f'\nPatient ID: {patient_info.name}\tday: {w._day}')
                    except AssertionError as e:
                        if abs(round(action[0].value, self._rounding_precision[arm]) - round(dose_info['Dose'][i], self._rounding_precision[arm])) > 2e-2 \
                                or dose_info['INR'][i] >= 5.0:
                            if arm[:-1] == 'pgpg' and i in [3, 4]:
                                a._dose = dose_info['Dose'][i]
                                continue
                            print(
                                f"{patient_info.name}, i: {i}, {action[0].value}, {dose_info['Dose'][i]}, {dose_info['INR'][i-1]}, {list(dose_info['Dose'][:i])}, {list(patient_info[:-3])}")
                            raised = True
                            break

        if raised:
            raise AssertionError


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
'''
ReilBase class
==============

The base class for reinforcement learning
'''

from __future__ import annotations

import bz2
import copy
import importlib
import logging
import pathlib
import time
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Union

import dill
from ruamel.yaml import YAML

import reil


Parsable = Union[OrderedDict[str, Any], Any]


class ReilBase:
    '''
    The base class of all classes in the `ReiL` package.
    '''

    def __init__(self,
                 name: Optional[str] = None,
                 path: Optional[pathlib.PurePath] = None,
                 logger_name: Optional[str] = None,
                 logger_level: Optional[int] = None,
                 logger_filename: Optional[str] = None,
                 persistent_attributes: Optional[List[str]] = None,
                 save_zipped: bool = False,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        name:
            An optional name for the instance that can be used to `save` the
            instance.

        path:
            An optional path to be used to `save` the instance.

        logger_name:
            Name of the `logger` that records logging messages.

        logger_level:
            Level of logging.

        logger_filename:
            An optional filename to be used by the logger.

        persistent_attributes:
            A list of attributes that should be preserved when loading an
            instance.

            For example, one might need to `load` an instance, but keep the
            name of the current instance.

            Example
            -------
            >>> instance = ReilBase(name='my_instance',
            ...                     persistent_attributes=['name'])
            >>> another_instance = ReilBase(name='another_instance')
            >>> another_instance.save('another_instance')
            >>> instance._name
            my_instance
            >>> another_instance._name
            another_instance
            >>> instance.load('another_instance')
            >>> instance._name
            my_instance

        kwargs:
            Any other attributes to set for the object.
            Note that `ReilBase` accepts any attribute and adds an
            underscore before its name.

            Example
            -------
            >>> instance = ReilBase(name='my_instance', my_attr='test')
            >>> instance._my_attr
            test
        '''
        self._name = name or self.__class__.__qualname__.lower()
        self._path = pathlib.PurePath(path or '.')
        self._save_zipped = save_zipped

        self._persistent_attributes = [
            '_' + p
            for p in (persistent_attributes or [])]

        self._logger_name = logger_name or __name__
        self._logger_level = logger_level or logging.WARNING
        self._logger_filename = logger_filename

        self._logger = logging.getLogger(self._logger_name)
        self._logger.setLevel(self._logger_level)
        if self._logger_filename is not None:
            self._logger.addHandler(logging.FileHandler(self._logger_filename))

        self.set_params(**kwargs)

    @classmethod
    def _empty_instance(cls):
        return cls()

    @classmethod
    def from_pickle(
            cls, filename: str,
            path: Optional[Union[pathlib.PurePath, str]] = None):
        '''
        Load a pickled instance.

        Arguments
        ---------
        filename:
            Name of the pickle file.

        path:
            Path of the pickle file.

        Returns
        -------
        :
            A `ReilBase` instance.
        '''
        instance = cls._empty_instance()
        instance._logger_name = __name__
        instance._logger_level = logging.WARNING
        instance._logger_filename = None
        instance._logger = logging.getLogger(instance._logger_name)
        instance._logger.setLevel(instance._logger_level)
        if instance._logger_filename is not None:
            instance._logger.addHandler(
                logging.FileHandler(instance._logger_filename))

        instance.load(filename=filename, path=path)
        return instance

    @classmethod
    def from_yaml_file(cls, node_reference: Tuple[str, ...],
                       filename: str,
                       path: Optional[Union[pathlib.PurePath, str]] = None):
        '''
        Create an instance based on a yaml file.

        Arguments
        ---------
        node_reference:
            A list of node names that determines the location of the
            specification in the yaml tree.

        filename:
            Name of the pickle file.

        path:
            Path of the pickle file.

        Returns
        -------
        :
            The generated instance.
        '''
        _path = pathlib.Path(path or '.')
        _filename = filename if filename.endswith((
            '.yaml', '.yml')) else f'{filename}.yaml'

        yaml = YAML()
        with open(_path / _filename, 'r') as f:
            yaml_output: OrderedDict[str, Any] = yaml.load(f)  # type: ignore

        temp_yaml = yaml_output
        for key in node_reference:
            temp_yaml = temp_yaml[key]

        return cls.parse_yaml(temp_yaml)

    @staticmethod
    def parse_yaml(data: Parsable) -> Parsable:  # noqa: C901
        '''
        Parse a yaml tree.

        This method reads a yaml tree and recursively creates objects specified
        by it.

        Arguments
        ---------
        data:
            A yaml tree data.

        Returns
        -------
        :
            Based on the tree, the method returns:

            * A python object, e.g. `int`, `float`, `str`.
            * A dictionary of arguments and their values to be fed to the
              'parse_yaml` caller.
            * An instance of an object derived from `ReilBase`.
        '''
        if isinstance(data, (int, float, str)):
            return data

        if 'eval' in data:
            args = {'reil': reil}
            if 'args' in data:
                args.update(ReilBase.parse_yaml(data['args']))
            # else:
            #     args = {}
            return eval(data['eval'], args)

        if len(data) == 1:
            k, v = next(iter(data.items()))
            result = ReilBase._create_component_from_yaml(k, v)
            if result is not None:
                return result

        args: Dict[str, Any] = {}
        for k, v in data.items():
            if isinstance(v, dict):
                v_obj = ReilBase.parse_yaml(data[k])
            elif isinstance(v, list):
                v_obj = [
                    ReilBase.parse_yaml(v_i) for v_i in v]  # type: ignore
            elif isinstance(v, str):
                if v.startswith('lambda'):
                    v_obj = eval(v, {})
                elif v.startswith('eval'):
                    v_obj = eval(v[4:], {})
                else:
                    v_obj = v
            else:
                v_obj = v

            args.update({k: v_obj})

        return args

    @staticmethod
    def _create_component_from_yaml(name: str, args: OrderedDict[str, Any]):
        '''
        Create a component from yaml data.

        This method attempts to import the `reil` class specified in `name`,
        parse arguments specified in `args` and create an instance of the
        class using the parsed arguments. If such class does not exist,
        `None` will be returned.

        Arguments
        ---------
        name:
            Name of the object to be created.

        args:
            A yaml tree section that contains arguments and values to create
            the object.

        Returns
        -------
        :
            The created object or `None`.
        '''
        temp = name.split('.')
        try:
            module = importlib.import_module('.'.join(temp[:-1]))
        except ValueError:
            return None

        f = getattr(module, temp[-1])
        if hasattr(f, 'parse_yaml'):
            result = f(**f.parse_yaml(args))
        else:
            result = f(**ReilBase.parse_yaml(args))

        return result

    def set_params(self, **params: Dict[str, Any]) -> None:
        '''
        set parameters to values.

        Arguments
        ---------
        params:
            A dictionary containing parameter names and their values.
        '''
        for key, value in params.items():
            self.__dict__[f'_{key}'] = value

    def load(self, filename: str,  # noqa: C901
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Load an object from a file.

        Arguments
        ---------
        filename:
            the name of the file to be loaded.

        path:
            the path in which the file is saved.

        Raises
        ------
        ValueError
            filename is not specified.

        RuntimeError
            Corrupted or inaccessible data file.
        '''
        _filename = (filename if filename.endswith(('.pkl', '.pbz2'))
                     else
                     f'{filename}{".pbz2" if self._save_zipped else ".pkl"}')
        full_path = pathlib.Path(path or self._path) / _filename

        data: Optional[Dict[str, Any]] = None
        for i in range(1, 6):
            try:
                if self._save_zipped:
                    with bz2.BZ2File(full_path, 'r') as f:
                        data = dill.load(f)  # type: ignore
                else:
                    with open(full_path, 'rb') as f:
                        data = dill.load(f)  # type: ignore
            except FileNotFoundError:
                raise
            except (EOFError, OSError):
                self._logger.info(
                    f'Attempt {i} failed to load '
                    f'{full_path}.')
                time.sleep(2)
            if data is not None:
                break

        if data is None:
            self._logger.exception(
                'Corrupted or inaccessible data file: '
                f'{full_path}')
            raise RuntimeError(
                f'Corrupted or inaccessible data file: '
                f'{full_path}')

        self._logger.info(
            'Changing the logger from '
            f'{self._logger_name} to {data["_logger_name"]}.')

        persistent_attributes = self._persistent_attributes + \
            ['_persistent_attributes']
        for key, value in data.items():
            if key not in persistent_attributes:
                self.__dict__[key] = value

        self._logger = logging.getLogger(self._logger_name)
        self._logger.setLevel(self._logger_level)
        if self._logger_filename is not None:
            self._logger.addHandler(
                logging.FileHandler(self._logger_filename))

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:
        '''
        Save the object to a file.

        Arguments
        ---------
        filename:
            the name of the file to be saved.

        path:
            the path in which the file should be saved.

        data_to_save:
            a list of variables that should be pickled. If omitted,
            the object is saved completely.

        Returns
        -------
        :
            a `Path` object to the location of the saved file and its name
            as `str`
        '''
        if data_to_save is None:
            data = self.__dict__
        else:
            data = {d: self.__dict__[d]
                    for d in list(data_to_save) + ['_name', '_path']}

        temp = None
        if '_logger' in data:
            temp = copy.deepcopy(self._logger)
            data.pop('_logger')

        _filename = filename or self._name
        _path = pathlib.Path(path or self._path)
        _path.mkdir(parents=True, exist_ok=True)

        if self._save_zipped:
            with bz2.BZ2File(_path / f'{_filename}.pbz2', 'w') as f:
                dill.dump(data, f, dill.HIGHEST_PROTOCOL)  # type: ignore
        else:
            with open(_path / f'{_filename}.pkl', 'wb+') as f:
                dill.dump(data, f, dill.HIGHEST_PROTOCOL)  # type: ignore

        if temp:
            self._logger = temp

        return pathlib.PurePath(_path), _filename

    def reset(self) -> None:
        ''' Reset the object.'''
        pass

    def __repr__(self) -> str:
        return self.__class__.__qualname__



# -*- coding: utf-8 -*-
'''
Agent class
===========

This `agent` class is the base class of all agent classes that can learn from
`history`.
'''

import pathlib
from typing import Any, Generator, Generic, Literal, Optional, Tuple, Union

from reil import stateful
from reil.agents.no_learn_agent import NoLearnAgent
from reil.datatypes.feature import FeatureArray
from reil.learners.learner import Learner, LabelType
from reil.utils.exploration_strategies import ExplorationStrategy

TrainingData = Tuple[Tuple[FeatureArray, ...], Tuple[LabelType, ...]]


class Agent(NoLearnAgent, Generic[LabelType]):
    '''
    The base class of all agent classes that learn from history.
    '''

    def __init__(self,
                 learner: Learner[LabelType],
                 exploration_strategy: ExplorationStrategy,
                 discount_factor: float = 1.0,
                 default_actions: Tuple[FeatureArray, ...] = (),
                 tie_breaker: Literal['first', 'last', 'random'] = 'random',
                 training_trigger: Literal[
                     'none', 'termination',
                     'state', 'action', 'reward'] = 'termination',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        learner:
            the `Learner` object that does the learning.

        exploration_strategy:
            an `ExplorationStrategy` object that determines
            whether the `action` should be exploratory or not for a given
            `state` at a given `iteration`.

        discount_factor:
            by what factor should future rewards be discounted?

        default_actions:
            a tuple of default actions.

        tie_breaker:
            how to choose the `action` if more than one is candidate
            to be chosen.

        training_trigger:
            When to learn from observations. This arguments is used in
            `observe` method to determine when `learn` method should be called.
            `none` avoids any call to `learn`; `state`, `action` and `reward`
            trigger the `learn` method after receiving their corresponding
            value; `termination` waits until `.close()` method of the generator
            is called.
        '''
        self._tie_breaker: Literal['first', 'last', 'random']

        super().__init__(default_actions, tie_breaker, **kwargs)

        self._learner: Learner[LabelType] = learner
        if not 0.0 <= discount_factor <= 1.0:
            self._logger.warning(
                f'{self.__class__.__qualname__} discount_factor should be in'
                f' [0.0, 1.0]. Got {discount_factor}. Set to 1.0.')
        self._discount_factor = min(discount_factor, 1.0)
        self._exploration_strategy = exploration_strategy
        self._training_trigger = training_trigger

    @classmethod
    def _empty_instance(cls):
        return cls(None, None)  # type: ignore

    def act(self,
            state: FeatureArray,
            subject_id: int,
            actions: Optional[Tuple[FeatureArray, ...]] = None,
            iteration: int = 0) -> FeatureArray:
        '''
        Return an action based on the given state.

        Arguments
        ---------
        state:
            the state for which the action should be returned.

        subject_id:
            the ID of the `subject` on which action should occur.

        actions:
            the set of possible actions to choose from.

        iteration:
            the iteration in which the agent is acting.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.

        Returns
        -------
        :
            the action
        '''
        if subject_id not in self._entity_list:
            raise ValueError(f'Subject with ID={subject_id} not found.')

        if (self._training_trigger != 'none' and
                self._exploration_strategy.explore(iteration)):
            possible_actions = actions or self._default_actions
            action = self._break_tie(
                possible_actions, self._tie_breaker)
        else:
            action = super().act(state=state, subject_id=subject_id,
                                 actions=actions, iteration=iteration)

        return action

    def reset(self):
        '''Reset the agent at the end of a learning iteration.'''
        super().reset()
        if self._training_trigger != 'none':
            self._learner.reset()

    def load(self, filename: str,
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Load an object from a file.

        Arguments
        ---------
        filename:
            the name of the file to be loaded.

        path:
            the path in which the file is saved.

        Raises
        ------
            ValueError
                Filename is not specified.
        '''
        _path = pathlib.Path(path or self._path)
        super().load(filename, _path)

        # when loading, self._learner is the object type, not an instance.
        self._learner = self._learner.from_pickle(  # type: ignore
            filename, _path / 'learner')

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:
        '''
        Save the object to a file.

        Arguments
        ---------
        filename:
            the name of the file to be saved.

        path:
            the path in which the file should be saved.

        data_to_save:
            a list of variables that should be pickled. If omitted,
            the `agent` is saved completely.

        Returns
        -------
        :
            a `Path` object to the location of the saved file and its name as
            `str`
        '''
        data = list(data_to_save or self.__dict__)
        save_learner = '_learner' in data
        if save_learner:
            temp, self._learner = (  # type: ignore
                self._learner, type(self._learner))

        _path = pathlib.Path(path or self._path)
        _filename = filename or self._name

        try:
            super().save(
                _filename, _path, data_to_save=tuple(data))
            if save_learner:
                self._learner = temp  # type: ignore
                self._learner.save(_filename, _path / 'learner')
        finally:
            if save_learner:
                self._learner = temp  # type: ignore

        return _path, _filename

    def _prepare_training(
            self, history: stateful.History) -> TrainingData[LabelType]:
        '''
        Use `history` to create the training set in the form of `X` and `y`
        vectors.

        Arguments
        ---------
        history:
            a `History` object from which the `agent` learns.

        Returns
        -------
        :
            a `TrainingData` object that contains `X` and 'y` vectors


        :meta public:
        '''
        raise NotImplementedError

    def learn(self, history: stateful.History) -> None:
        '''
        Learn using history.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` whose history is being used for learning.

        next_state:
            The new `state` of the `subject` after taking `agent`'s action.
            Some methods
        '''
        training_data: TrainingData[Any] = (), ()
        if history is not None:
            training_data = self._prepare_training(history)

        X, Y = training_data
        if X:
            self._learner.learn(X, Y)

    def observe(self, subject_id: int, stat_name: Optional[str],  # noqa: C901
                ) -> Generator[Union[FeatureArray, None], Any, None]:
        '''
        Create a generator to interact with the subject (`subject_id`).
        Extends `NoLearnAgent.observe`.

        This method creates a generator for `subject_id` that
        receives `state`, yields `action` and receives `reward`
        until it is closed. When `.close()` is called on the generator,
        `statistics` are calculated.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` on which action happened.

        stat_name:
            The name of the `statistic` that should be computed at the end of
            each trajectory.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.
        '''
        if subject_id not in self._entity_list:
            raise ValueError(f'Subject with ID={subject_id} not found.')

        trigger = self._training_trigger
        learn_on_state = trigger == 'state'
        learn_on_action = trigger == 'action'
        learn_on_reward = trigger == 'reward'
        learn_on_termination = trigger == 'termination'

        history: stateful.History = []
        new_observation = stateful.Observation()
        while True:
            try:
                new_observation = stateful.Observation()
                temp = yield
                new_observation.state = temp['state']
                actions: Tuple[FeatureArray, ...] = temp['actions']
                iteration: int = temp['iteration']

                if learn_on_state:
                    self.learn([history[-1], new_observation])

                if actions is not None:
                    new_observation.action = self.act(
                        state=new_observation.state,  # type: ignore
                        subject_id=subject_id,
                        actions=actions, iteration=iteration)

                    if learn_on_action:
                        self.learn([history[-1], new_observation])

                    new_observation.reward = (yield new_observation.action)

                    history.append(new_observation)

                    if learn_on_reward:
                        self.learn(history[-2:])
                else:
                    yield

            except GeneratorExit:
                if new_observation.reward is None:  # terminated early!
                    history.append(new_observation)

                if learn_on_termination:
                    self.learn(history)

                if stat_name is not None:
                    self.statistic.append(stat_name, subject_id)

                return



# -*- coding: utf-8 -*-  pylint: disable=undefined-variable
'''
HealthSubject class
===================

This `HealthSubject` class implements interaction with patients.
'''

from typing import Any, List, Optional, Tuple, Union

from reil.datatypes.feature import (MISSING, Feature, FeatureArray,
                                    FeatureGenerator)
from reil.healthcare.patient import Patient
from reil.subjects.subject import Subject


class HealthSubject(Subject):
    '''
    A HealthSubject subject class with a patient.
    '''

    def __init__(
            self,
            patient: Patient,
            measurement_name: str,
            measurement_range: Tuple[float, float],
            dose_range: Tuple[float, float],
            interval_range: Tuple[int, int],
            max_day: int,
            **kwargs: Any):
        '''
        Arguments
        ---------
        patient:
            A patient object that generates new patients and models
            interaction between dose and INR.

        measurement_name:
            Name of the measured value after dose administration, e.g. BGL,
            INR, etc.

        dose_range:
            A tuple that shows the minimum and maximum amount of dose of
            the medication.

        interval_range:
            A tuple that shows the minimum and maximum duration of each dosing
            decision.

        measurement_range:
            A tuple that shows the minimum and maximum possible values of each
            measurement.

        max_day:
            Maximum duration of each trial.
        '''

        super().__init__(max_agent_count=1, **kwargs)

        self._patient = patient
        if not self._patient:
            return

        self._dose_range = dose_range
        self._interval_range = interval_range
        self._measurement_range = measurement_range
        self._measurement_name = measurement_name

        self._max_day = max_day

        self.feature_gen_set = {
            name: FeatureGenerator.numerical(
                name=name, lower=lower, upper=upper)
            for name, lower, upper in (
                (f'{self._measurement_name}_history',
                 *self._measurement_range),
                (f'daily_{self._measurement_name}_history',
                 *self._measurement_range),
                ('dose_history', *self._dose_range),
                ('daily_dose_history', *self._dose_range),
                ('interval_history', *self._interval_range),
                ('dose', *self._dose_range),
                ('interval', *self._interval_range),
                ('day', 0, self._max_day - 1)
            )
        }

        self._generate_state_defs()
        self._generate_reward_defs()
        self._generate_statistic_defs()
        self._generate_action_defs()

        self.reset()

    def _generate_state_defs(self):
        self.state.add_definition('day',
                                  ('day', {}))

    def _generate_reward_defs(self):
        self.reward.add_definition(
            'no_reward', lambda _: 0.0, 'day')

    def _generate_statistic_defs(self):
        pass

    def _generate_action_defs(self):
        pass

    @classmethod
    def _empty_instance(cls):
        return cls(None, None)  # type: ignore

    @staticmethod
    def generate_dose_values(
            min_dose: float = 0.0,
            max_dose: float = 15.0,
            dose_increment: float = 0.5) -> List[float]:

        return list(min_dose + x * dose_increment
                    for x in range(
                        int((max_dose - min_dose)/dose_increment) + 1))

    @staticmethod
    def generate_interval_values(
            min_interval: int = 1,
            max_interval: int = 28,
            interval_increment: int = 1) -> List[int]:

        return list(range(min_interval, max_interval, interval_increment))

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        return self._day >= self._max_day

    def take_effect(
            self, action: FeatureArray, _id: int = 0) -> None:
        Subject.take_effect(self, action, _id)
        action_temp = action.value
        current_dose = float(action_temp['dose'])
        current_interval = min(int(action_temp['interval']),
                               self._max_day - self._day)

        measurements_temp = self._patient.model(
            dose={
                i: current_dose
                for i in range(self._day, self._day + current_interval)
            },
            measurement_days=list(
                range(self._day + 1, self._day + current_interval + 1)
            )
        )[self._measurement_name]

        self._decision_points_dose_history[self._decision_points_index] = \
            current_dose
        self._decision_points_interval_history[self._decision_points_index] = \
            current_interval
        self._decision_points_index += 1

        day_temp = self._day
        self._day += current_interval

        self._full_dose_history[day_temp:self._day] = \
            [current_dose] * current_interval
        self._full_measurement_history[day_temp +
                                       1:self._day + 1] = measurements_temp

        self._decision_points_measurement_history[
            self._decision_points_index] = \
            self._full_measurement_history[self._day]

    def reset(self) -> None:
        Subject.reset(self)
        self._patient.generate()

        self._day: int = 0
        self._full_measurement_history = [0.0] * self._max_day
        self._full_dose_history = [0.0] * self._max_day
        self._decision_points_measurement_history = [0.0] * (self._max_day + 1)
        self._decision_points_dose_history = [0.0] * self._max_day
        self._decision_points_interval_history: List[int] = [1] * self._max_day
        self._decision_points_index: int = 0

        self._full_measurement_history[0] = self._patient.model(
            measurement_days=[0])[self._measurement_name][-1]
        self._decision_points_measurement_history[0] = \
            self._full_measurement_history[0]

    def _numerical_sub_comp(self, name: str):
        return self._patient.feature_set[name]

    def _categorical_sub_comp(self, name: str, missing: bool = False):
        if missing:
            self._patient.feature_gen_set[name](MISSING)

        return self._patient.feature_set[name]

    def _get_history(
            self, list_name: str, length: int
    ) -> Union[Feature[List[float]], Feature[List[int]]]:
        if length == 0:
            raise ValueError(
                'length should be a positive integer, or '
                '-1 for full length output.')

        filler: Union[float, int]
        _list: Union[List[float], List[int]]
        if list_name == f'{self._measurement_name}_history':
            _list = self._decision_points_measurement_history
            index = self._decision_points_index + 1
            filler = 0.0
        elif list_name == f'daily_{self._measurement_name}_history':
            _list = self._full_measurement_history
            index = self._day + 1
            filler = 0.0
        elif list_name == 'dose_history':
            _list = self._decision_points_dose_history
            index = self._decision_points_index
            filler = 0.0
        elif list_name == 'daily_dose_history':
            _list = self._full_dose_history
            index = self._day
            filler = 0.0
        elif list_name == 'interval_history':
            _list = self._decision_points_interval_history
            index = self._decision_points_index
            filler = int(1)
        else:
            raise ValueError(f'Unknown list_name: {list_name}.')

        if length == -1:
            result = _list[:index]
        else:
            if length > index:
                i1, i2 = length - index, 0
            else:
                i1, i2 = 0, index-length
            result = [filler] * i1 + _list[i2:index]  # type: ignore

        return self.feature_gen_set[list_name](result)  # type: ignore

    def _sub_comp_dose_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._get_history('dose_history', length)  # type: ignore

    def _sub_comp_measurement_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._get_history(  # type: ignore
            f'{self._measurement_name}_history', length)

    def _sub_comp_interval_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[int]]:
        return self._get_history('interval_history', length)  # type: ignore

    def _sub_comp_day(self, _id: int, **kwargs: Any) -> Feature[int]:
        return self.feature_gen_set['day'](  # type: ignore
            value=self._day if 0 <= self._day < self._max_day else None)

    def _sub_comp_daily_dose_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._get_history('daily_dose_history', length)  # type: ignore

    def _sub_comp_daily_measurement_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._get_history(  # type: ignore
            f'daily_{self._measurement_name}_history', length)

    def __repr__(self) -> str:
        try:
            temp = ', '.join(''.join(
                (str(k), ': ',
                 ('{:4.2f}' if v.is_numerical else '{}').format(v.value)))
                for k, v in self._patient.feature_set.items())
        except (AttributeError, ValueError, KeyError):
            temp = ''

        return (f'{self.__class__.__qualname__} [{temp}]')



# -*- coding: utf-8 -*-
'''
InteractionProtocol class
=========================

A datatype that accepts initial value and feature generator, and generates
new values. This datatype uses `Entity` to specify an `agent` or a `subject`.
'''
import dataclasses
from typing import Literal, Optional, Tuple


@dataclasses.dataclass(frozen=True)
class Entity:
    '''
    The datatype to specify an `agent` or a `subject`.
    Used in `InteractionProtocol`.
    '''
    name: str
    demon_name: Optional[str] = None
    statistic_name: Optional[str] = None
    groupby: Optional[Tuple[str, ...]] = None
    aggregators: Optional[Tuple[str, ...]] = None

    def __post_init__(self):
        if self.groupby is not None:
            self.__dict__['groupby'] = tuple(self.groupby)
        if self.aggregators is not None:
            self.__dict__['aggregators'] = tuple(self.aggregators)


@dataclasses.dataclass
class InteractionProtocol:
    '''
    The datatype to specify how an `agent` should interact with a `subject` in
    an `environment`.
    '''
    agent: Entity
    subject: Entity
    state_name: str
    action_name: str
    reward_name: str
    n: int
    unit: Literal['interaction', 'instance', 'iteration']



# -*- coding: utf-8 -*-
'''
HambergPKPD class
=================

A warfarin PK/PD model proposed by Hamberg et al. (2007).
DOI: 10.1038/sj.clpt.6100084
'''
import math
from typing import (Any, Callable, Dict, Final, Iterable, List, NamedTuple,
                    NewType, Union)

import numpy as np
from reil.datatypes.feature import Feature
from reil.healthcare.mathematical_models.health_math_model import \
    HealthMathModel

Day = NewType('Day', float)
Hour = NewType('Hour', int)
dT = NewType('dT', int)


class DoseEffect(NamedTuple):
    dose: float
    Cs: Callable[[Iterable[dT]], List[float]]


class HambergPKPD(HealthMathModel):
    '''
    Hamberg PK/PD model for warfarin.
    '''
    _per_hour: int = 1

    # Hamberg et al. (2007) - Table 2
    # CL_s: Apparent oral clearance
    _CL_s_1_1: Final = 0.314  # (l/h) CL_s for *1/*1 of a typical 71-yr patient
    _CL_s_genotypes: Final = {  # Effect of genotypes diff. from *1/*1 on CL_s
        '*1/*1': 0.0,
        '*1/*2': 0.315,
        '*1/*3': 0.453,
        '*2/*2': 0.722,
        '*2/*3': 0.690,
        '*3/*3': 0.852}
    _CL_s_age: Final = 0.0091  # Effect of age on CL_s centered around 71 years
    _V1: Final = 13.8  # (l) Apparent central volume of distribution
    _k_aS: Final = 2.0  # (1/hr) Absorption rate constant
    _Q: Final = 0.131    # (l/h) Apparent intercompartmental clearance
    _V2: Final = 6.59  # (l) Apparent peripheral volume of distribution

    _omega_CL_s: Final = 0.310  # Interindividual variability for CL_s
    _omega_V1: Final = 0.262  # Interindividual variability for V1
    _omega_V2: Final = 0.991  # Interindividual variability for V2

    _sigma_s: Final = 0.0908  # Residual error after single dose
    _sigma_ss: Final = 0.301  # Residual error after steady-state dose

    # Hamberg et al. (2007) - Table 4
    _E_max: Final = 1.0
    _gamma: Final = 0.424
    _EC_50_GG: Final = 4.61  # (mg/l) EC_50 for VKORC1 genotype G/G
    _EC_50_GA: Final = 3.02  # (mg/l) EC_50 for VKORC1 genotype G/A
    _EC_50_AA: Final = 2.20  # (mg/l) EC_50 for VKORC1 genotype A/A
    _MTT_1: Final = 11.6  # (h) Mean Transit Time
    _MTT_2: Final = 120  # (h) Mean Transit Time
    _lambda: Final = 3.61

    _omega_MTT_1: Final = 0.141  # Interindividual variability for MTT_1
    _omega_MTT_2: Final = 1.020  # Interindividual variability for MTT_2
    _omega_EC_50: Final = 0.409  # Interindividual variability for EC_50
    _sigma_INR: Final = 0.0325  # Residual error for INR

    # Hamberg et al. (2007) - Misc.
    _INR_max: Final = 20.0  # page 538

    def __init__(
            self, randomized: bool = True,
            cache_size: Day = Day(30)) -> None:
        """
        Arguments
        ---------
        randomized:
            Whether to have random effects in patient response to warfarin.

        cache_size:
            Size of the cache used to store pre-computed values needed for
            INR computation.
        """
        self._randomized = randomized
        self._cache_size = math.ceil(cache_size)
        self._last_computed_day: Day = Day(0)
        self._cached_cs: Dict[float, List[float]] = {}

    def setup(self, **arguments: Feature[Any]) -> None:
        '''
        Set up the model.

        Arguments
        ---------
        arguments:
            `Feature` instances required to setup the model.

        Notes
        -----
        This model requires `age`, `CYP2C9`, `MTT_1`, `MTT_2`, `EC_50`,
        `cyp_1_1`, `V1`, and `V2`. The genotype of `VKORC1` is not directly
        used in this implementation. Instead, one should use it to generate
        `EC_50`. See `WarfarinPatient` class.

        Raises
        ------
        ValueError:
            `CYP2C9` is not one of the acceptable values:
            *1/*1, *1/*2, *1/*3, *2/*2, *2/*3, *3/*3
        '''
        # Note: In Hamberg et al. (2007), BASE_i is the measured baseline INR
        # for patients, but Ravvaz fixed it to 1.
        self._baseINR = 1.0    # Ravvaz source code

        age = float(arguments['age'].value)  # type: ignore
        CYP2C9 = str(arguments['CYP2C9'].value)
        MTT_1 = float(arguments['MTT_1'].value)  # type: ignore
        MTT_2 = float(arguments['MTT_2'].value)  # type: ignore
        V1 = float(arguments['V1'].value)  # type: ignore
        V2 = float(arguments['V2'].value)  # type: ignore
        EC_50 = float(arguments['EC_50'].value)  # type: ignore
        CL_S_cyp_1_1 = float(arguments['CL_S_cyp_1_1'].value)  # type: ignore

        if CYP2C9 not in self._CL_s_genotypes:
            raise ValueError('The CYP2C9 genotype not recognized!')

        # Note:
        # Hamberg et al. (2007) says "CLS was reduced with increasing age,
        # decreasing by approximately 9% per decade". However, it does not say
        # whether it applies to younger than 71 or not. Ravvaz assumed that it
        # does not, but Figure 3 in Hamberg et al. (2007) shows a 50-year-old
        # patient has a lower INR than a 70-year-old. So, it applies!
        CL_s = CL_S_cyp_1_1 * (
            1.0 - (self._CL_s_age * (age - 71.0))
        ) * (1 - self._CL_s_genotypes[CYP2C9])

        # -------- Implementation of the two compartment model ------------
        # Adapted from Ravvaz source code in R
        # Similar to "Equations to use for time points when absorption can
        # occur and compartment A is unsaturated" available at
        # http://www.rsc.org/suppdata/c7/md/c7md00586e/c7md00586e1.pdf
        # There, compartments B and C are equivalent of compartments 1 and 2
        # here. Also, k_1...k_4 are K_aS, k12, k21, k10, respectively.
        # Note that $x_B^0$ and $x_C^0$ terms are zero as described in that
        # document.

        # bioavilability fraction 0-1 (from: "Applied Pharmacokinetics &
        # Pharmacodynamics 4th edition, p.717", some other references)
        F = 0.9

        k12 = self._Q / V1  # Central to peripheral distribution constant
        k21 = self._Q / V2  # Peripheral to central distribution constant
        k10 = CL_s / V1  # Elimination rate constant

        b = k10 + k21 + k12
        c = k10 * k21

        # Alpha: distribution phase slope (-alpha)
        # Beta: elimination phase slope (-beta)
        self._alpha = (b + math.sqrt(b ** 2 - 4*c)) / 2
        self._beta = (b - math.sqrt(b ** 2 - 4*c)) / 2

        # Note: Here we halved the value for KaF_2V1, because half of the
        # warfarin is S, and only S affects the INR.
        kaF_2V1 = (self._k_aS * F / 2) / V1

        # $C_s$ = c_{\k_1}\exp{-k_1 t}
        #       + c_{\alpha}\exp{-\alpha t}
        #       + c_{\beta}\exp{-\beta t}
        self._coef_alpha = (
            (k21 - self._alpha)
            / ((self._k_aS - self._alpha)*(self._beta - self._alpha))
        ) * kaF_2V1
        self._coef_beta = (
            (k21 - self._beta)
            / ((self._k_aS - self._beta)*(self._alpha - self._beta))
        ) * kaF_2V1
        self._coef_k_a = (
            (k21 - self._k_aS)
            / ((self._k_aS - self._alpha)*(self._k_aS - self._beta))
        ) * kaF_2V1

        # ---- End of the implementation of the two compartment model -------

        # Note: According to Hamberg et al. (2007) pp. 538, $ktr_1=1/MTT_1$
        # However, Ravvaz set it to $6/MTT_1$. It must be because we have
        # 6 compartment amounts, and total $MTT_1$ for the whole chain is
        # 11.6 h (Figure 2), so each should take $\frac{1}{6} MTT_1$
        ktr1 = 6.0/MTT_1  # (1/hours)
        ktr2 = 1.0/MTT_2  # (1/hours)
        self._ktr = np.array([0.0] + [ktr1] * 6 + [0.0, ktr2])
        self._EC_50_gamma = EC_50 ** self._gamma

        self._dose_records: Dict[Day, DoseEffect] = {}
        cs_size = self._cache_size * 24 * self._per_hour
        self._total_cs = np.array([0.0] * cs_size)  # hourly
        # self._computed_INRs = [0.0] * (self._cache_size + 1)  # daily
        self._computed_INRs: Dict[Day, float] = {}  # daily
        self._err_list: List[List[float]] = []  # hourly
        self._err_ss_list: List[List[float]] = []  # hourly
        self._exp_e_INR_list: List[List[float]] = []  # daily
        self._last_computed_day = Day(0)

        temp_cs_generator = self._CS_function_generator(dT(0), 1.0)
        self._cached_cs = {
            1.0: temp_cs_generator(range(cs_size))}  # type: ignore

        self._A = np.array([0.0] + [1.0] * 8)
        # self._computed_INRs[0] = self._INR(self._A, Day(0))
        self._computed_INRs[Day(0)] = self._INR(self._A, Day(0))

    def run(self, **inputs: Any) -> Dict[str, Any]:
        '''
        Run the model.

        Arguments
        ---------
        inputs:
            - A dictionary called "dose" with days for each dose as keys and
              the amount of dose as values.
            - A list called "measurement_days" that shows INRs of which days
              should be returned.

        Returns
        -------
        :
            A dictionary with keyword "INR" and a list of doses for the
            specified days.
        '''
        self.dose = inputs.get('dose', {})

        if days := inputs.get('measurement_days'):
            return {'INR': self.INR(days)}

        return {'INR': {}}

    @property
    def dose(self) -> Dict[Day, float]:
        '''
        Return doses for each day.

        Returns
        -------
        :
            A dictionary with days as keys and doses as values.
        '''
        return {t: info.dose
                for t, info in self._dose_records.items()}

    @dose.setter
    def dose(self, dose: Dict[Day, float]) -> None:
        '''
        Add warfarin doses at the specified days.

        Arguments
        ---------
        dose:
            A dictionary with days as keys and doses as values.
        '''
        # if a dose is added ealier in the list, INRs should be updated all
        # together because the history of "A" array is not kept.
        try:
            if self._last_computed_day > min(dose.keys()):
                self._last_computed_day = Day(0)
        except ValueError:  # no doses
            pass

        for day, _dose in dose.items():
            if _dose != 0.0:
                dt = dT(math.ceil(day * 24 * self._per_hour))
                if day in self._dose_records:
                    # TODO: Implement!
                    raise NotImplementedError

                self._dose_records[day] = DoseEffect(
                    _dose, self._CS_function_generator(dt, _dose))

                self._total_cs += np.array(self._dose_records[day].Cs(range(
                    self._cache_size * 24 * self._per_hour)))  # type: ignore

    def INR(self, measurement_days: Union[Day, List[Day]]) -> List[float]:
        '''
        Compute INR values for the specified days.

        Arguments
        ---------
        measurement_days:
            One of a list of all days for which INR should be computed.

        Returns
        -------
        :
            A list of INRs for the specified days.
        '''
        days: List[Day]

        days = (measurement_days if hasattr(measurement_days, '__iter__')
                else [measurement_days])  # type: ignore

        not_computed_days = set(days).difference(self._computed_INRs)
        if (not_computed_days and
                min(not_computed_days) < self._last_computed_day):
            self._last_computed_day = Day(0)
            self._computed_INRs = {}
            not_computed_days = days

        if self._last_computed_day == 0:
            self._A = np.array([0.0] + [1.0] * 8)  # type: ignore

        stop_points = [self._last_computed_day] + list(not_computed_days)
        for d1, d2 in zip(stop_points[:-1], stop_points[1:]):
            delta_Ts = 24 * self._per_hour
            for dt in range(int(d1 * delta_Ts), int(d2 * delta_Ts)):
                self._A[0] = self._A[7] = self._inflow(dt)  # type: ignore
                self._A[1:] += self._ktr[1:] * \
                    (self._A[0:-1] - self._A[1:]) / self._per_hour

            self._computed_INRs[d2] = self._INR(self._A, Day(int(d2)))

        self._last_computed_day = stop_points[-1]

        return [self._computed_INRs[i] for i in days]

    def _CS_function_generator(
            self, dt_dose: dT, dose: float
    ) -> Callable[[Iterable[dT]], List[float]]:
        '''
        Generate a Cs function.

        Arguments
        ---------
        dt_dose:
            The time in which the dose is administered.

        dose:
            The value of the dose administered.

        Returns
        -------
        :
            A function that gets the time and returns that time's
            warfarin concentration.

        Notes
        -----
        To speed up the process, the generated function uses a pre-computed
        cache of concentrations and only computes the concentration
        if the requested day is beyond the cached range.
        '''
        if dose == 1.0:
            cached_cs_temp = []
        else:
            if dose not in self._cached_cs:
                self._cached_cs[dose] = [dose * cs
                                         for cs in self._cached_cs[1.0]]
            cached_cs_temp = self._cached_cs[dose]

        def Cs(dts: Iterable[dT]) -> List[float]:
            '''
            Get delta_t list and return the warfarin concentration of
            those times.

            Arguments
            ---------
            dts:
                The times for which concentration value is needed.

            Returns
            -------
            :
                Warfarin concentration
            '''
            max_diff = len(cached_cs_temp)
            coef_alpha = self._coef_alpha
            coef_beta = self._coef_beta
            coef_k_a = self._coef_k_a
            alpha = self._alpha
            beta = self._beta
            k_aS = self._k_aS

            # For hour_diff == 0, the Cs equation itself is zero, so included
            # it in the main if to avoid unnecessary computation.
            return [
                0.0 if (dt_diff := dt - dt_dose) <= 0
                else (
                    cached_cs_temp[dt_diff] if dt_diff < max_diff
                    else (
                        coef_alpha *
                        math.exp(-alpha * dt_diff / self._per_hour) +
                        coef_beta *
                        math.exp(-beta * dt_diff / self._per_hour) +
                        coef_k_a *
                        math.exp(-k_aS * dt_diff / self._per_hour)
                    ) * dose
                ) for dt in dts]

        return Cs

    def _err(self, dt: dT, ss: bool = False) -> float:
        '''
        Generate error term for the requested day.

        Arguments
        ---------
        dt:
            The time for which the error is requested.

        ss:
            Whether the error is for the steady-state case. For single dose
            it should be `False`, and for multiple doses, it should be `True`.

        Returns
        -------
        :
            The error value.

        Notes
        -----
        To speed up the process and generate reproducible results in each run,
        the errors are cached in batches.
        For each call of the function, the cached error is returned. If
        the `hour` is beyond the cached range, a new range of error values
        are generated and added to the cache.
        Also, error is per hour. So, for any fraction of hour, the same
        value will be returned.
        '''
        if self._randomized:
            h = dt // self._per_hour
            hourly_cache_size = self._cache_size * 24
            index_0 = h // hourly_cache_size
            index_1 = h % hourly_cache_size
            e_list = self._err_ss_list if ss else self._err_list
            try:
                return e_list[index_0][index_1]
            except IndexError:
                missing_rows = index_0 - len(e_list) + 1
                stdev = self._sigma_ss if ss else self._sigma_s
                for _ in range(missing_rows):
                    e_list.append(np.exp(np.random.normal(  # type:ignore
                        0, stdev, hourly_cache_size)))

            return e_list[index_0][index_1]
        else:
            return 1.0

    def _exp_e_INR(self, d: Day) -> float:
        '''
        Generate exp(error) term of INR for the requested day.

        Arguments
        ---------
        d:
            The day for which the error is requested.

        Returns
        -------
        :
            The error value.

        Notes
        -----
            error is generated per day. So, if the given `d` is fractional,
            it will be truncated and the respective error is returned.

            To speed up the process and generate reproducible results in each
            run, the errors are cached in batches.
            For each call of the function, the cached error is returned. If
            the `day` is beyond the cached range, a new range of error values
            are generated and added to the cache.
        '''
        _d = int(d)
        if self._randomized:
            index_0 = _d // self._cache_size
            index_1 = _d % self._cache_size
            try:
                return self._exp_e_INR_list[index_0][index_1]
            except IndexError:
                missing_rows = index_0 - len(self._exp_e_INR_list) + 1
                for _ in range(missing_rows):
                    self._exp_e_INR_list.append(
                        np.exp(np.random.normal(  # type:ignore
                            0, self._sigma_INR,
                            self._cache_size)))

            return self._exp_e_INR_list[index_0][index_1]

        else:
            return 1.0

    def _inflow(self, t: dT) -> float:
        '''
        Compute the warfarin concentration that enters the two compartments
        in the PK/PD model.

        Arguments
        ---------
        t:
            The time for which the input is requested.
            t = Hours * _per_hour + delta_t

        Returns
        -------
        :
            The input value.

        Notes
        -----
        To speed up the process, total concentration is being cached for a
        number of days. For days beyond this range, concentration values are
        computed and used on each call.
        '''
        try:
            Cs = self._total_cs[t]
        except IndexError:
            Cs = sum(v.Cs([t])[0]
                     for v in self._dose_records.values())

        Cs_gamma = (Cs * self._err(t, t > 0)) ** self._gamma
        inflow = 1 - (
            (self._E_max * Cs_gamma) / (self._EC_50_gamma + Cs_gamma))

        return inflow

    def _INR(self, A: List[float], d: Day) -> float:
        '''
        Compute the INR on day `d`.

        Arguments
        ---------

        d:
            The day for which the input is requested.

        Returns
        -------
        :
            The INR value.

        Notes
        -----
        To speed up the process, total concentration is being cached for a
        number of days. For days beyond this range, concentration values are
        computed and used on each call.
        '''

        # Note: we defined `A` in such a way to compute changes in `A`s
        # easier. In our implementation, `A8` is the `A7` in Hamberg et al.
        return (self._baseINR +
                (self._INR_max * ((1 - A[6]*A[8]) ** self._lambda))
                ) * self._exp_e_INR(d)



#%%
import weka.core.jvm as jvm
import weka.core.converters as converters
from weka.filters import Filter
from weka.clusterers import Clusterer
import weka.core.serialization as serialization
from weka.core.dataset import Instances

class WekaClusterer:
    def __init__(self, model_filename=None, data_filename=None, columns_to_drop=[]):
        if not jvm.started:
            jvm.start()
        self._data_filename = data_filename
        self._columns_to_drop = columns_to_drop
        self._model_filename = model_filename
        if self._model_filename is not None:
            self.load_model()
        elif self._data_filename is not None:
            self.load_data()

    def load_data(self, data_filename=None, columns_to_drop=None):
        self._data_filename = data_filename if data_filename is not None else self._data_filename
        self._columns_to_drop = columns_to_drop if columns_to_drop is not None else self._columns_to_drop
        data = converters.load_any_file(self._data_filename)

        if self._columns_to_drop == []:
            self._data = data
        else:
            remove = Filter(classname="weka.filters.unsupervised.attribute.Remove", options=['-R', ', '.join(str(i) for i in self._columns_to_drop)])
            remove.inputformat(data)     # let the filter know about the type of data to filter
            self._data = remove.filter(data)   # filter the data

    def cluster(self, clustering_method='EM', number_of_clusters=-1, **options):
        try:
            classname = {'em': 'weka.clusterers.EM',
                         'kmeans': 'weka.clusterers.SimpleKMeans'}[clustering_method.lower()]
        except KeyError:
            raise KeyError(f'Method {clustering_method} not found!')
        
        self._clusterer = Clusterer(classname=classname, options=['-N', str(number_of_clusters)] + options)
        self._clusterer.build_clusterer(self._data)

        return self._clusterer

    def assign_label(self, data, distribution=False, **kwargs):
        results = [0]*len(data)
        if distribution:
            for index, inst in enumerate(data):
                cl = clusterer.cluster_instance(inst)  # 0-based cluster index
                dist = clusterer.distribution_for_instance(inst)   # cluster membership distribution
                results[index] = (cl, dist)
        else:
            for index, inst in enumerate(data):
                results[index] = self._clusterer.cluster_instance(inst)  # 0-based cluster index

        return results

    def save_model(self, model_filename=None, include_instances=True):
        filename = self._model_filename if model_filename is None else model_filename
        if include_instances:
            serialization.write_all(filename, [self._clusterer, Instances.template_instances(self._data)])
        else:
            serialization.write(filename, self._clusterer)

    def load_model(self, model_filename=None, has_instances=True):
        filename = self._model_filename if model_filename is None else model_filename
        if has_instances:
            objects = serialization.read_all(filename)
            self._clusterer = Clusterer(jobject=objects[0])
            self._instances = Instances(jobject=objects[1])
        else:
            self._clusterer = Clusterer(jobject=serialization.read(filename))

    def __del__(self):
        jvm.stop()
    

#%%
if __name__ == "__main__":
    filename = 'data_for_clustering_dose_only.csv'
    test = WekaClusterer(data_filename=filename, columns_to_drop=[1])
    test.cluster(number_of_clusters=4)
    test.save_model(model_filename='weka_output.model')
    test.load_model(model_filename='weka_output.model')
    test.assign_label(test._data)


from __future__ import annotations

import dataclasses
import functools
from typing import Any, List, NamedTuple, Optional, Tuple

from reil.datatypes.feature import FeatureArray
from reil.utils.functions import in_range, interpolate, square_dist

# SOME THOUGHTS!
# Lookahead is not something a subject can/should do!
# The environment should take a copy of the subject, apply some policy (random,
# current policy, fixed, etc.) for a number of steps (ReilFunction.length) for
# a number of times (sample size), then provide the outcome to the ReilFunction
# and finally add that to the reward from the subject!


class Arguments(NamedTuple):
    y: str
    x: Optional[str] = None


@dataclasses.dataclass
class ReilFunction:
    name: str
    arguments: Arguments
    length: int = -1
    multiplier: float = 1.0
    retrospective: bool = True
    interpolate: bool = True

    def __post_init__(self):
        if not isinstance(self.arguments, Arguments):  # type: ignore
            self.arguments = Arguments(*self.arguments)  # type: ignore

    def __call__(self, args: FeatureArray) -> float:
        temp = args.value
        y = temp[self.arguments.y]
        x = temp[self.arguments.x] if self.arguments.x is not None else None

        fn = [functools.partial(self._no_retro_no_inter, y),
              functools.partial(self._no_retro_inter, y, x),
              functools.partial(self._retro_no_inter, y),
              functools.partial(self._retro_inter, y, x)
              ][self.retrospective*2 + self.interpolate]

        try:
            result = self.multiplier * fn()
        except NotImplementedError:
            result = self.multiplier * self._default_function(y, x)

        return result

    def _retro_inter(self, y: List[Any], x: List[Any]) -> float:
        raise NotImplementedError

    def _retro_no_inter(self, y: List[Any]) -> float:
        raise NotImplementedError

    def _no_retro_inter(self, y: List[Any], x: List[Any]) -> float:
        raise NotImplementedError

    def _no_retro_no_inter(self, y: List[Any]) -> float:
        raise NotImplementedError

    def _default_function(
            self, y: List[Any], x: Optional[List[Any]] = None) -> float:
        raise NotImplementedError


@dataclasses.dataclass
class NormalizedSquareDistance(ReilFunction):
    center: float = 0.0
    band_width: float = 1.0
    exclude_first: bool = False

    def _default_function(
            self, y: List[Any], x: Optional[List[Any]] = None) -> float:
        _x = x or [1] * (len(y) - 1)

        if len(y) != len(_x) + 1:
            raise ValueError(
                'y should have exactly one item more than x.')

        if not self.exclude_first:
            _x = [1] + _x
            _y = [0.0] + y
        else:
            _y = y

        result = sum(square_dist(
            self.center, interpolate(_y[i], _y[i+1], _x[i]))
            for i in range(len(_x)))

        # normalize
        result *= (2.0 / self.band_width) ** 2

        return result


@dataclasses.dataclass
class PercentInRange(ReilFunction):
    acceptable_range: Tuple[float, float] = (0.0, 1.0)
    exclude_first: bool = False

    def _default_function(
            self, y: List[Any], x: Optional[List[Any]] = None) -> float:
        _x = x or [1] * (len(y) - 1)
        if len(y) != len(_x) + 1:
            raise ValueError(
                'y should have exactly one item more than x.')

        if not self.exclude_first:
            _x = [1] + _x
            _y = [0.0] + y
        else:
            _y = y

        result = sum(
            in_range(
                self.acceptable_range,
                interpolate(_y[i], _y[i+1], _x[i]))
            for i in range(len(_x)))

        total_intervals = sum(_x)

        return result / total_intervals


# TODO: not implemented yet!
@dataclasses.dataclass
class Delta(ReilFunction):
    '''
    Get changes in the series.

    available `op`s:
        count: counts the number of change points in y.
        sum: sum of value changes
        average: average value change

    available `interpolation_method`s:
        linear
        post: y = y[i] at x[i]
        pre: y = y[i] at x[i-1]
    '''
    exclude_first: bool = False
    op: str = 'count'
    interpolation_method: str = 'linear'

# def _default_function(
#         self, y: List[Any], x: Optional[List[Any]] = None) -> float:
#     if self.op == 'count':
#         result = sum(yi != y[i+1]
#                     for i, yi in enumerate(y[:-1]))

#     return result


class Functions:
    @staticmethod
    def TTR(INRs: List[float],
            intervals: Optional[List[int]] = None,
            exclude_first: bool = False,
            INR_range: Tuple[float, float] = (2.0, 3.0)) -> float:
        if intervals is None:
            temp = INRs[1:] if exclude_first else INRs
            result = sum((1 if INR_range[0] <= INRi <= INR_range[1] else 0
                          for INRi in temp)
                         ) / len(temp)
        else:
            if len(INRs) != len(intervals) + 1:
                raise ValueError(
                    'INRs should have exactly one item more than intervals.')

            result = 0.0
            for i, current_interval in enumerate(intervals):
                result += sum(
                    1
                    if INR_range[0] <=
                    (INRs[i] + (INRs[i+1] - INRs[i])/current_interval*j)
                    <= INR_range[1]
                    else 0
                    for j in range(1, current_interval + 1))

            total_intervals = sum(intervals)
            if not exclude_first:
                result += Functions.TTR([INRs[0]])
                total_intervals += 1

            result /= total_intervals

        return result

    @staticmethod
    def dose_change_count(dose_list: List[float],
                          intervals: Optional[List[int]] = None) -> int:
        # assuming dose is fixed during each interval
        return sum(x != dose_list[i+1]
                   for i, x in enumerate(dose_list[:-1]))

    @staticmethod
    def delta_dose(dose_list: List[float],
                   intervals: Optional[List[int]] = None) -> float:
        # assuming dose is fixed during each interval
        return sum(abs(x-dose_list[i+1])
                   for i, x in enumerate(dose_list[:-1]))

    @staticmethod
    def total_dose(dose_list: List[float],
                   intervals: Optional[List[int]] = None) -> float:
        if intervals is None:
            result = sum(dose_list)
        else:
            if len(dose_list) != len(intervals):
                raise ValueError(
                    'dose_list and intervals should '
                    'have the same number of items.')

            result = sum(dose*interval
                         for dose, interval in zip(dose_list, intervals))

        return result

    @staticmethod
    def average_dose(dose_list: List[float],
                     intervals: Optional[List[int]] = None) -> float:
        total_dose = Functions.total_dose(dose_list, intervals)
        total_interval = len(
            dose_list) if intervals is None else sum(intervals)

        return total_dose / total_interval

    @staticmethod
    def normalized_square_dist(
            INRs: List[float],
            intervals: Optional[List[int]] = None,
            exclude_first: bool = False,
            INR_range: Tuple[float, float] = (2.0, 3.0)) -> float:

        INR_mid = sum(INR_range) / 2.0

        _intervals = intervals or [1] * (len(INRs) - 1)

        if len(INRs) != len(_intervals) + 1:
            raise ValueError(
                'INRs should have exactly one item more than intervals.')

        if not exclude_first:
            _intervals = [1] + _intervals
            _INRs = [0.0] + INRs
        else:
            _INRs = INRs

        result = sum(
            square_dist(
                INR_mid,
                interpolate(_INRs[i], _INRs[i+1], _intervals[i]))
            for i in range(len(_intervals)))

        # normalize
        result *= (2.0 / (INR_range[1] - INR_range[0])) ** 2

        return result



# -*- coding: utf-8 -*-
'''
Lenzini class
=============

Lenzini warfarin dosing protocol based on `Lenzini et al. (2010)
<https://doi.org/10.1038/clpt.2010.13>`_
'''

import functools
from math import exp, log, sqrt
from typing import Any, Dict, Literal, Tuple

import reil.healthcare.dosing_protocols.dosing_protocol as dp


class Lenzini(dp.DosingProtocol):
    '''
    Lenzini warfarin dosing protocol based on `Lenzini et al. (2010)
    <https://doi.org/10.1038/clpt.2010.13>`_
    '''

    def __init__(self,
                 method: Literal['pharmacogenetic',
                                 'clinical'] = 'pharmacogenetic') -> None:
        '''
        Arguments
        ---------
        method:
            One of 'pharmacogenetic', 'clinical'.
        '''
        if method.lower() == 'clinical':
            self._method = self.clinical
        elif method.lower() in ['pg', 'pharmacogenetic', 'default']:
            self._method = self.pg
        else:
            raise ValueError(
                f'Unknown method {method}.\n'
                'Acceptable methods are clinical and pharmacogenetic.')

    def prescribe(self,  # noqa: C901
                  patient: Dict[str, Any],
                  additional_info: dp.AdditionalInfo
                  ) -> Tuple[dp.DosingDecision, dp.AdditionalInfo]:
        dose = self._method(patient)

        return dp.DosingDecision(dose), additional_info

    @staticmethod
    def clinical(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using clinical Lenzini formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - BSA (m^2)
            - INR_history (The most recent value will be used.)
            - target_INR
            - race (Only 'Black' changes the dose.)
            - stroke ('Yes', 'No')
            - diabetes ('Yes', 'No')
            - amiodarone ('Yes', 'No')
            - fluvastatin ('Yes', 'No')
            - dose_history (A list of previous doses at decision points)
            - interval_history (A list of previous intervals between decision
                points)
        Returns
        -------
        :
            The adjusted dose

        Notes
        -----
        If BSA is not provided, weight (lb) and height (in) is used to
        calculate BSA using Mosteller method.

        target_INR is assumed 2.5 if not provided.
        '''
        if not 4 <= patient['day'] <= 5:
            raise ValueError('Lenzini can only be used on days 4 and 5.')

        if (bsa := patient.get('BSA')) is None:
            try:
                bsa = sqrt(patient['height'] * 2.54 *
                           patient['weight'] * 0.454 / 3600)
            except KeyError:
                raise KeyError(
                    'Either `BSA` in m^2, or `weight` in lb and `height` in in'
                    'should be provided.')

        age = patient['age']
        ln_INR = log(patient['INR_history'][-1])  # Natural log

        target_INR = patient.get('target_INR', 2.5)

        african_american = patient.get('race') == 'Black'
        stroke = patient.get('stroke') == 'Yes'
        diabetes = patient.get('diabetes') == 'Yes'
        amiodarone = patient.get('amiodarone') == 'Yes'
        fluvastatin = patient.get('fluvastatin') == 'Yes'

        dose_history = patient['dose_history']
        interval_history = patient['interval_history']
        all_doses = functools.reduce(
            lambda x, y: x+y,
            ([dose_history[-i]]*interval_history[-i]
             for i in range(len(interval_history), 0, -1)))

        # Since on day 4, we don't have dose-4, a zero is padded to avoid
        # producing error in the run time.
        dose_4, dose_3, dose_2 = ([0] + all_doses)[-4:-1]

        dose = exp(2.81602
                   - 0.76679 * ln_INR
                   - 0.00590 * age
                   + 0.27815 * target_INR
                   - 0.16759 * diabetes
                   + 0.17675 * bsa
                   - 0.22844 * stroke
                   - 0.25487 * fluvastatin
                   + 0.07123 * african_american
                   - 0.11137 * amiodarone
                   + 0.03471 * dose_2
                   + 0.03047 * dose_3
                   + 0.01929 * dose_4
                   ) / 7

        return dose

    @staticmethod
    def pg(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using pharmacogenetic Lenzini formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - BSA (m^2)
            - INR_history (The most recent value will be used.)
            - VKORC1
            - CYP2C9
            - target_INR
            - race (Only 'Black' changes the dose.)
            - stroke ('Yes', 'No')
            - diabetes ('Yes', 'No')
            - amiodarone ('Yes', 'No')
            - fluvastatin ('Yes', 'No')
            - dose_history (All previous doses administered.)
        Returns
        -------
        :
            The adjusted dose

        Notes
        -----
        If BSA is not provided, weight (lb) and height (in) is used to
        calculate BSA using Mosteller method.

        target_INR is assumed 2.5 if not provided.
        '''
        if not 4 <= patient['day'] <= 5:
            raise ValueError('Lenzini can only be used on days 4 and 5.')

        if (bsa := patient.get('BSA')) is None:
            try:
                bsa = sqrt(patient['height'] * 2.54 *
                           patient['weight'] * 0.454 / 3600)
            except KeyError:
                raise KeyError(
                    'Either `BSA` in m^2, or `weight` in lb and `height` in in'
                    'should be provided.')

        age = patient['age']
        ln_INR = log(patient['INR_history'][-1])  # Natural log

        # Both Lenzini et al. and Ravvaz et al. have VKORC1 G/A, in the formula
        # but Lenzini mentions that it is coded based on the number of "A"s
        # A/A -> 2, G/A -> 1, G/G -> 0
        vkorc1: str = patient['VKORC1']
        vkorc1_a = vkorc1.count('A')

        cyp2c9: str = patient['CYP2C9']
        cyp2c9_2 = cyp2c9.count('2')
        cyp2c9_3 = cyp2c9.count('3')

        target_INR = patient.get('target_INR', 2.5)

        african_american = patient.get('race') == 'Black'
        stroke = patient.get('stroke') == 'Yes'
        diabetes = patient.get('diabetes') == 'Yes'
        amiodarone = patient.get('amiodarone') == 'Yes'
        fluvastatin = patient.get('fluvastatin') == 'Yes'

        # Since on day 4, we don't have dose-4, a zero is padded to avoid
        # producing error in the run time.
        dose_4, dose_3, dose_2 = ([0] + patient['dose_history'])[-4:-1]

        dose = exp(3.10894
                   - 0.00767 * age
                   - 0.51611 * ln_INR
                   - 0.23032 * vkorc1_a
                   - 0.14745 * cyp2c9_2
                   - 0.30770 * cyp2c9_3
                   + 0.24597 * bsa
                   + 0.26729 * target_INR
                   - 0.09644 * african_american
                   - 0.20590 * stroke
                   - 0.11216 * diabetes
                   - 0.10350 * amiodarone
                   - 0.19275 * fluvastatin
                   + 0.01690 * dose_2
                   + 0.02018 * dose_3
                   + 0.01065 * dose_4
                   ) / 7

        return dose



# -*- coding: utf-8 -*-
'''
MNKGame class
==============

This class emulates mnk game.
'''


import random
from typing import Any, Dict, Optional, Tuple, Union

from reil.datatypes.feature import FeatureArray, FeatureGenerator
from reil.subjects.subject import Subject
from reil.utils.mnkboard import MNKBoard


class MNKGame(MNKBoard, Subject):
    '''
    Build an m-by-n board (using `MNKBoard` class) in which p players can play.
    Winner is the player who can put `k` pieces in on row, column, or diagonal.
    '''
    # _board is a row vector. (row, column) and index start from 0
    # _board_status: None: no winner yet,
    #                1..players: winner,
    #                0: stall,
    #               -1: illegal board

    def __init__(self, m: int, n: int, k: int, players: int, **kwargs: Any):
        '''
        Arguments
        ---------
        m:
            The number of rows.

        n:
            The number of columns.

        k:
            The winning criterion, i.e. the number of cells in one row,
            column, or diagonal that a player needs to capture to win the game.

        players:
            The number of players.
        '''
        self._board_status = None
        self.set_params(**kwargs)
        Subject.__init__(self, **kwargs)
        MNKBoard.__init__(self, m=m, n=n, k=k, players=players,
                          can_recapture=False, **kwargs)
        self._action_gen = FeatureGenerator.numerical(
            name='square', lower=0, upper=len(self._board)-1)

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        return self._board_status is not None

    def possible_actions(
            self, _id: Optional[int] = None) -> Tuple[FeatureArray, ...]:
        return tuple(FeatureArray(self._action_gen(v))
                     for v in self.get_action_set())

    def take_effect(self, action: FeatureArray, _id: int) -> None:
        '''
        Set a piece for the given player on the board.

        Arguments
        ---------
        action:
            The location in which the piece is placed.

        _id:
            ID of the player who sets the piece.
        '''
        Subject.take_effect(self, action, _id)
        self.set_piece(_id, index=int(action.value['square']))

    def _default_reward_definition(
            self, _id: Optional[int] = None) -> int:
        if self._board_status is None:
            r = 0
        elif self._board_status == _id:
            r = 1
        elif self._board_status > 0:
            r = -1
        else:
            r = 0

        return r

    def reset(self):
        '''Clear the board and update board_status.'''
        Subject.reset(self)
        MNKBoard.reset(self)
        self._board_status = None

    def set_piece(self, player: int, index: Optional[int] = None,
                  row: Optional[int] = None, column: Optional[int] = None
                  ) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        player:
            ID of the player whose piece will be set on the board.

        row:
            The row on which the piece should be placed.

        column:
            The column on which the piece should be placed.

        index:
            Where to put the piece. Index starts from 0 and assumes the board
            to be a list.

        Raises
        ------
        ValueError:
            Wrong player ID.

        ValueError:
            index is out of range.

        Notes
        -----
        Either `index` or `row` and `column` should be used. If both are used,
        `row` and `column` is used.
        '''
        super().set_piece(
            player=player, index=index, row=row, column=column)  # type: ignore
        if self._board_status is None:
            self._board_status = self._update_board_status(
                player=player, index=index, row=row, column=column)
        elif self._board_status > 0:
            self._board_status = -1

    def _update_board_status(self, player: int,  # noqa: C901
                             index: Optional[int] = None,
                             row: Optional[int] = None,
                             column: Optional[int] = None
                             ) -> Union[int, None]:
        # player wins: player | doesn't win: None | draw: 0
        '''
        Get a player and the location of the latest change and try to find
        a sequence of length k of the specified player.

        Arguments
        ---------
        player:
            ID of the player whose piece will be set on the board.

        row:
            The row on which the piece should be placed.

        column:
            The column on which the piece should be placed.

        index:
            Where to put the piece. Index starts from 0 and assumes the board
            to be a list.

        Return
        ------
        :
            0: sequence not found and the board is full (stall)
            player: sequence found (win)
            None: sequence not found and the board is not full (ongoing)

        Notes
        -----
        Either `index` or `row` and `column` should be used. If both are used,
        `row` and `column` is used.
        '''
        if index is None:
            if row is None or column is None:
                raise TypeError('No (row, column) or index found')
            else:
                r = row
                c = column
        else:
            r = index // self._n
            c = index % self._n

        ul_r = max(r - self._k + 1, 0)
        ul_c = max(c - self._k + 1, 0)
        lr_r = min(r + self._k, self._m - 1)
        lr_c = min(c + self._k, self._n - 1)
        m = self.list_to_matrix(self._board, self._m, self._n)

        # Vertical sequence
        pointer = ul_r
        counter: int = 0
        while pointer <= lr_r:
            counter = (counter + 1) * (m[pointer][c] == player)
            if counter == self._k:
                return player
            pointer += 1

        # Horizontal sequence
        pointer = ul_c
        counter = 0
        while pointer <= lr_c:
            counter = (counter + 1) * (m[r][pointer] == player)
            if counter == self._k:
                return player
            pointer += 1

        # Diagonal \
        min_d = min(r - ul_r, c - ul_c)
        pointer_r = r - min_d
        pointer_c = c - min_d
        counter = 0
        while pointer_r <= lr_r and pointer_c <= lr_c:
            counter = (counter + 1) * (m[pointer_r][pointer_c] == player)
            if counter == self._k:
                return player
            pointer_r += 1
            pointer_c += 1

        # Diagonal /
        min_d = min(r - ul_r, lr_c - c)
        pointer_r = r - min_d
        pointer_c = c + min_d
        counter = 0
        while pointer_r <= lr_r and pointer_c >= ul_c:
            counter = (counter + 1) * (m[pointer_r][pointer_c] == player)
            if counter == self._k:
                return player
            pointer_r += 1
            pointer_c -= 1

        if min(self._board) > 0:
            return 0

        return None

    def __repr__(self):
        return self.__class__.__qualname__


if __name__ == '__main__':
    board = MNKGame(m=3, n=3, k=3, players=2)
    player: Dict[str, int] = {}
    p = 0
    player['P1'] = board.register('P1')
    player['P2'] = board.register('P2')
    while not board.is_terminated():
        current_player = ['P1', 'P2'][p]
        print(p, current_player)
        actions = board.possible_actions(player[current_player])
        board.take_effect(random.choice(actions), player[current_player])
        print(f'{board}\n',
              board.reward('default', player['P1']),
              board.reward('default', player['P2']))
        p = (p + 1) % 2



from logging import warning
import math
import random
from typing import Callable, List, Literal
import unittest

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from reil.datatypes.feature import Feature, FeatureGenerator
from reil.healthcare.mathematical_models import HambergPKPD
from reil.utils.functions import (random_categorical, random_lognormal,
                                  random_normal_truncated, random_uniform)


class testHambergPKPD(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        random.seed(1000)

    def test_replicate_fig_3(self) -> None:
        # Replicating Hamberg et al. (2007) Figure 3

        # keys are subplots of the Fig 3, the tuple represents each
        # patient's 'age', 'CYP2C9', and 'VKORC1' along with 'dose'
        # and the last 'INR'. Since the paper does not have exact
        # values, we ran the model as best as we could to replicate
        # the paper, and then recorded those numbers for reference.
        # This allows us to test any changes in the PKPD
        # implementation, but is no guarantee that these numbers are
        # what they should be. Also, this code saves a chart that
        # can be studied and compared with Fig 3. The zigzag
        # behavior in Fig 3 is the result of hourly INR measurement.
        # We only measure INR daily, hence the smooth lines.
        patients = {
            (0, 0): [(50.0, '*1/*1', 'G/G', 10.0, 2.41),
                     (50.0, '*1/*2', 'G/G', 10.0, 3.00),
                     (50.0, '*1/*3', 'G/G', 10.0, 3.42),
                     (50.0, '*2/*2', 'G/G', 10.0, 5.03),
                     (50.0, '*2/*3', 'G/G', 10.0, 4.73),
                     (50.0, '*3/*3', 'G/G', 10.0, 6.90)],
            (0, 1): [(50.0, '*1/*1', 'G/G', 10.0, 2.41),
                     (50.0, '*1/*1', 'G/A', 10.0, 3.05),
                     (50.0, '*1/*1', 'A/A', 10.0, 3.66)],
            (1, 0): [(50.0, '*1/*1', 'G/G', 10.0, 2.41),
                     (70.0, '*1/*1', 'G/G', 10.0, 2.65),
                     (90.0, '*1/*1', 'G/G', 10.0, 2.97)],
            (1, 1): [(50.0, '*1/*1', 'G/G', 10.0, 2.41),
                     (50.0, '*1/*1', 'A/A', 10.0, 3.66),
                     (50.0, '*1/*3', 'G/A', 10.0, 4.35),
                     (50.0, '*3/*3', 'G/G', 10.0, 6.90)]}

        other_features = {
            'MTT_1': Feature(name='MTT_1', value=HambergPKPD._MTT_1),
            'MTT_2': Feature(name='MTT_2', value=HambergPKPD._MTT_2),
            'CL_S_cyp_1_1': Feature(
                name='CL_S_cyp_1_1', value=HambergPKPD._CL_s_1_1),
            'V1': Feature(name='V1', value=HambergPKPD._V1),
            'V2': Feature(name='V2', value=HambergPKPD._V2)
        }

        _, axes = plt.subplots(
            2, 2, figsize=(10, 10), sharex=True, sharey=True)

        h = HambergPKPD(cache_size=60, randomized=False)  # type: ignore
        h._per_hour = 1
        errors = ''
        for (i, j), info in patients.items():
            for age, cyp, vkor, dose, inr in info:
                if vkor == 'G/G':
                    v = HambergPKPD._EC_50_GG
                elif vkor in ('G/A', 'A/G'):
                    v = HambergPKPD._EC_50_GA
                else:  # 'A/A'
                    v = HambergPKPD._EC_50_AA

                other_features['EC_50'] = Feature(name='EC_50', value=v)

                h.setup(
                    age=Feature(name='age', value=age),
                    CYP2C9=Feature(name='CYP2C9', value=cyp),
                    VKORC1=Feature(name='VKORC1', value=vkor),
                    **other_features)
                h.run(
                    dose={d: dose for d in range(60)},
                    measurement_days=range(61))
                try:
                    self.assertAlmostEqual(
                        h._computed_INRs[60], inr, 2)  # type: ignore
                except AssertionError as e:
                    errors += f'\n({age:4.2f}, {cyp}, {vkor})\t{e}'

                sns.lineplot(data=h._computed_INRs, ax=axes[i][j])
                axes[i][j].set_ylabel('INR', fontsize=12)
                axes[i][j].set_ylim((0, 8))
                axes[i][j].set_xlim((0, 59))
                axes[i][j].set_xlabel('Time (days)', fontsize=12)
                axes[i][j].set_xticks([0, 10, 20, 30, 40, 50])
                for x in (2, 2.5, 3, 4, 5, 6, 7):
                    axes[i][j].axhline(x, ls='--')

        plt.tight_layout()
        plt.savefig('replicated_Hamberg_et_al_2007_Fig_3.png')
        if errors:
            raise AssertionError(errors)

    def test_replicate_fig_4a(self) -> None:
        data = testHambergPKPD.fig_4_generator(
            100, random_pkpd=False)
        data = data[data.t > 0]
        data['t_w_jitter'] = data.t  # + data.x_jitter
        ax = sns.scatterplot(
            data=data, x='t_w_jitter', y='Cs', size=1, legend=False)
        print(data.groupby(['t'])['Cs'].quantile(0.025),
              data.groupby(['t'])['Cs'].median(),
              data.groupby(['t'])['Cs'].quantile(0.975),
              )
        sns.lineplot(data=data.groupby(['t'])['Cs'].quantile(0.025), ax=ax)
        sns.lineplot(data=data.groupby(['t'])['Cs'].median(), ax=ax)
        sns.lineplot(data=data.groupby(['t'])['Cs'].quantile(0.975), ax=ax)
        ax.set_ylim((0.0, 0.8))
        ax.set_xlabel('Time (h)')
        ax.set_ylabel('Concentration (mg/l)')
        plt.savefig('replicated_Hamberg_et_al_2007_Fig_4a.png')

    def test_replicate_fig_4c(self) -> None:
        data = testHambergPKPD.fig_4_generator(
            300, INR=True,
            random_parameters=True,
            random_demographic=True,
            random_pkpd=True,
            measurement_days=[0, 0.5, 1.5, 2.5]
            )
        data['t_w_jitter'] = data.t + data.x_jitter
        ax = sns.scatterplot(
            data=data, x='t_w_jitter', y='INR', size=1, legend=False)
        # style='run')
        print(data.groupby(['t'])['INR'].quantile(0.025),
              data.groupby(['t'])['INR'].median(),
              data.groupby(['t'])['INR'].quantile(0.975),
              )
        sns.lineplot(data=data.groupby(['t'])['INR'].quantile(0.025), ax=ax)
        sns.lineplot(data=data.groupby(['t'])['INR'].median(), ax=ax)
        sns.lineplot(data=data.groupby(['t'])['INR'].quantile(0.975), ax=ax)
        ax.set_ylim((0.8, 2.0))
        ax.set_xlabel('Time (h)')
        ax.set_ylabel('INR')
        plt.savefig('replicated_Hamberg_et_al_2007_Fig_4c.png')

    @staticmethod
    def fig_4_generator(
        n,
        days: int = 60,
        measurement_days: List[float] = [0, 0.5, 1.5, 2.5],
        INR: bool = False,
        random_parameters: bool = True,
        random_demographic: bool = True,
        random_pkpd: bool = True,
        random_gen: Callable[[FeatureGenerator], float] = random_lognormal,
        data_source: Literal[
            'Study I', 'Study II', 'Total', 'Ravvaz'] = 'Total'
    ) -> pd.DataFrame:
        if data_source == 'Study I':
            age_info = dict(lower=46.0, upper=87.0, mean=72.0)
            CYP2C9_info = dict(probabilities=(
                0.596, 0.193, 0.158, 0.053, 0.0, 0.0))
            VKORC1_info = dict(probabilities=(0.39, 0.41, 0.20))
        elif data_source == 'Study II':
            age_info = dict(lower=22.0, upper=84.0, mean=70.0)
            CYP2C9_info = dict(probabilities=(
                0.580, 0.161, 0.172, 0.022, 0.043, 0.022))
            warning.warn(
                'VKORC1 is not available for Study II. Total is used instead.')
            VKORC1_info = dict(probabilities=(0.39, 0.41, 0.20))
        elif data_source == 'Total':
            age_info = dict(lower=22.0, upper=87.0, mean=71.0)
            CYP2C9_info = dict(probabilities=(
                0.587, 0.173, 0.167, 0.033, 0.027, 0.013))
            VKORC1_info = dict(probabilities=(0.39, 0.41, 0.20))
        elif data_source == 'Ravvaz':
            age_info = dict(lower=18.0, upper=100.0, mean=67.30, stdev=13.43)
            CYP2C9_info = dict(probabilities=(
                0.6739, 0.1486, 0.0925, 0.0651, 0.0197, 2e-4))
            VKORC1_info = dict(probabilities=(0.3837, 0.4418, 0.1745))
        else:
            raise ValueError

        # VKORC1 and EC_50 are not important in Cs (They affect INR)
        feature_gen_set = {
            'age': FeatureGenerator.numerical(
                name='age',
                **age_info,
                generator=(random_uniform if data_source != 'Ravvaz'
                           else random_normal_truncated),
                randomized=random_demographic),
            'CYP2C9': FeatureGenerator.categorical(
                name='CYP2C9',
                categories=('*1/*1', '*1/*2', '*1/*3',
                            '*2/*2', '*2/*3', '*3/*3'),
                **CYP2C9_info,
                generator=random_categorical,
                randomized=random_demographic,
                allow_missing=True),
            'VKORC1': FeatureGenerator.categorical(
                name='VKORC1',
                categories=('G/G', 'G/A', 'A/A'),
                **VKORC1_info,
                generator=random_categorical,
                randomized=random_demographic,
                allow_missing=True),

            'CL_S_cyp_1_1': FeatureGenerator.numerical(
                name='CL_S_cyp_1_1',
                mean=math.log(HambergPKPD._CL_s_1_1),
                stdev=math.sqrt(HambergPKPD._omega_CL_s),
                generator=random_gen,
                randomized=random_parameters),
            'MTT_1': FeatureGenerator.numerical(
                name='MTT_1',
                mean=math.log(HambergPKPD._MTT_1),
                stdev=math.sqrt(HambergPKPD._omega_MTT_1),
                generator=random_gen,
                randomized=random_parameters),
            'MTT_2': FeatureGenerator.numerical(
                name='MTT_2',
                mean=math.log(HambergPKPD._MTT_2),
                stdev=math.sqrt(HambergPKPD._omega_MTT_2),
                generator=random_gen,
                randomized=random_parameters),
            'V1': FeatureGenerator.numerical(
                name='V1',
                mean=math.log(HambergPKPD._V1),
                stdev=math.sqrt(HambergPKPD._omega_V1),
                generator=random_gen,
                randomized=random_parameters),
            'V2': FeatureGenerator.numerical(
                name='V2',
                mean=math.log(HambergPKPD._V2),
                stdev=math.sqrt(HambergPKPD._omega_V2),
                generator=random_gen,
                randomized=random_parameters),
        }

        concentrations = []
        for i in range(n):
            h = HambergPKPD(
                cache_size=days, randomized=random_pkpd)  # type: ignore
            features = {key: value() for key, value in feature_gen_set.items()}
            vkor = features['VKORC1'].value
            if vkor == 'G/G':
                v = math.log(HambergPKPD._EC_50_GG)
            elif vkor in ('G/A', 'A/G'):
                v = math.log(HambergPKPD._EC_50_GA)
            else:  # 'A/A'
                v = math.log(HambergPKPD._EC_50_AA)

            features['EC_50'] = FeatureGenerator.numerical(
                name='EC_50',
                mean=v,
                stdev=HambergPKPD._omega_EC_50,
                generator=random_gen,
                randomized=random_parameters)()

            h.setup(**features)
            INR_values = h.run(
                dose={0: 10.0},
                measurement_days=measurement_days if INR else [])['INR']
            if not INR:
                INR_values = [0.0] * len(measurement_days)
            # sns.lineplot(y=h._total_cs, x=range(len(h._total_cs)))
            concentrations.extend([
                (i, t * 24,
                 h._total_cs[int(t * 24)]
                 * h._err(int(t * 24), False),  # type: ignore
                 INR_values[j], (random.random() - 0.5) * 2)
                for j, t in enumerate(measurement_days)])
        data = pd.DataFrame(
            concentrations, columns=['run', 't', 'Cs', 'INR', 'x_jitter'])

        return data


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
'''
datatypes module for reinforcement learning
===========================================

This module contains datatypes used in `reil`

Submodules
----------
buffers:
    A module that contains different types of `Buffers` used in `reil`
    package.

Classes
-------
Feature:
    A datatype that accepts initial value and feature generator and
    generates new values.

Entity:
    A datatype to specify `agent` or `subject` information. Used in
    `InteractionProtocol`.

InteractionProtocol:
    A datatype to specifies how an `agent` and a `subject`
    interact in an `environment`.

FeatureArray:
    The main datatype used to communicate `state`s, `action`s, and `reward`s,
    between objects in `reil`. `FeatureArray` is basically a dictionary that
    contains instances of `Feature`.

Feature:
    An immutable dataclass that accepts name, value, and is_numerical.

Categorical:
    A factory class that creates `Feature` objects with categories.

Numerical:
    A factory class that creates `Feature` objects with lower and upper
    bounds.

PrimaryComponent:
    A datatype that is being used mostly by children of `Stateful` to include
    a `PrimaryComponent`, e.g. a state. It allows defining different
    definitions for the component, and call the instance to calculate them.

SecondayComponent:
    A datatype that is being used mostly by children of `Subject` to include
    a `SecondayComponent`, e.g. a statistic or a reward. It allows defining
    different definitions for the component, and call the instance to calculate
    them.
'''
from . import buffers  # noqa: W601
from .feature import Feature, FeatureArray, FeatureGenerator  # noqa: W601
from .components import (PrimaryComponent, SecondayComponent,  # noqa: W601
                         Statistic)
from .interaction_protocol import Entity, InteractionProtocol  # noqa: W601
from .mock_statistic import MockStatistic  # noqa: W601



from .dosing_protocol import (AdditionalInfo, DosingDecision,  # noqa: W0611
                              DosingProtocol)
from .three_phase_dosing_protocol import ThreePhaseDosingProtocol  # noqa: W0611



# -*- coding: utf-8 -*-
'''
A Reinforcement Learning Module for Python
==========================================

This module provides a framework for training and test of different
reinforcement learning methods.

Modules
-------
agents
    Entities that act on one or more subject via an environment and observing
    the reward. `Agents` can learn or just be actors.

subjects
    Entities with an internal state that get one or more
    `agents`' action via an `environment` and return new state and reward.

environments
    Classes that connect `agents` and `subjects`, and simulate their
    interactions.

learners
A set of learning techniques used as the learner of an `agent`.

stats
    Compute statistics.

reilbase
    Base class for all `reil` classes.

stateful
    Based class for all stateful objects.

datatypes
    All custom datatypes used in `reil`.

utils
    Classes and functions that are utilities used in `reil`.

legacy
    All classes that are no longer supported.

@author: Sadjad Anzabi Zadeh (sadjad-anzabizadeh@uiowa.edu)
'''

from ._version import get_versions
__version__ = get_versions()['version']  # type: ignore
del get_versions



from typing import (Any, Callable, DefaultDict, Dict, List, Optional, Tuple,
                    Union)

import pandas as pd
from reil.datatypes.components import PrimaryComponent
from reil.datatypes.feature import FeatureArray
from reil.stateful import Stateful


class MockStatistic:
    '''
    A component that mocks `Statistic` class, and uses another object's
    `Statistic` methods, except for `append` and `aggregate`.
    '''

    def __init__(self, obj: Stateful) -> None:
        '''

        Parameters
        ----------
        obj:
            The object that provides actual `Statistic` capabilities.
        '''
        self._obj = obj
        self._history: Dict[
            int,
            List[Tuple[FeatureArray, float]]] = DefaultDict(list)
        self._history_none: List[Tuple[FeatureArray, float]] = []

    def set_object(self, obj: Stateful) -> None:
        self._obj = obj

    def enable(self) -> None:
        return self._obj.statistic.enable()

    def disable(self) -> None:
        return self._obj.statistic.disable()

    def set_primary_component(
            self,
            primary_component: PrimaryComponent) -> None:
        return self._obj.statistic.set_primary_component(primary_component)

    def add_definition(
            self,
            name: str,
            fn: Callable[..., Any],
            stat_component: str,
            aggregation_component: str) -> None:
        return self._obj.statistic.add_definition(
            name, fn, stat_component, aggregation_component)

    def default(self, _id: Optional[int] = None) -> Tuple[FeatureArray, float]:
        return self._obj.statistic.default(_id)

    def __call__(
            self,
            name: str,
            _id: Optional[int] = None
    ) -> Union[Tuple[FeatureArray, float], None]:
        return self._obj.statistic.__call__(name, _id)

    def append(
            self,
            name: str,
            _id: Optional[int] = None) -> None:
        '''
        Generate the stat and append it to the history.

        Arguments
        ---------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        s = self._obj.statistic.__call__(name, _id)
        # print(s[0].value, s[1])
        if s is not None:
            if _id is None:
                self._history_none.append(s)
            else:
                self._history[_id].append(s)

    def aggregate(
            self,
            aggregators: Optional[Tuple[str, ...]] = None,
            groupby: Optional[Tuple[str, ...]] = None,
            _id: Optional[int] = None,
            reset_history: bool = False):
        temp = self._history_none if _id is None else self._history[_id]
        if not temp:
            return None

        df = pd.DataFrame({'instance_id': i,  # type: ignore
                           **x[0].value,
                           'value': x[1]}
                          for i, x in enumerate(temp))
        temp_group_by = ['instance_id'] if groupby is None else list(groupby)
        grouped_df = df.groupby(temp_group_by)

        def no_change(x: Any) -> Any:
            return x

        result: pd.DataFrame = grouped_df['value'].agg(  # type: ignore
            aggregators or no_change)

        if reset_history:
            self._history: Dict[
                int, List[Tuple[FeatureArray, float]]] = DefaultDict(list)
            self._history_none: List[Tuple[FeatureArray, float]] = []

        return result



# -*- coding: utf-8 -*-
'''
Buffer class
============

The base class for all buffers in `reil`.
'''

from typing import (Dict, Generic, List, Literal, Optional, Tuple, TypeVar,
                    Union)

import numpy as np
from reil import reilbase

T1 = TypeVar('T1')
T2 = TypeVar('T2')

PickModes = Literal['all', 'random', 'recent', 'old']


class Buffer(reilbase.ReilBase, Generic[T1, T2]):
    '''
    The base class for all buffers in `reil`.
    '''

    def __init__(
            self,
            buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None,
            clear_buffer: Optional[bool] = None):
        '''
        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        buffer_names:
            A list containing the names of buffer queues.

        pick_mode:
            The default mode to pick items from the list.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.
        '''
        self._buffer: Union[None, Dict[str, Union[List[T1], List[T2]]]]
        self._buffer_size: Optional[int] = None
        self._buffer_names: Optional[List[str]] = None
        self._pick_mode: Optional[PickModes] = None
        self._clear_buffer: Optional[bool] = None
        self._buffer_index: int = -1
        self._count: int = 0

        self.setup(
            buffer_size=buffer_size, buffer_names=buffer_names,
            pick_mode=pick_mode, clear_buffer=clear_buffer)

    def setup(
            self, buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None,
            clear_buffer: Optional[bool] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        buffer_names:
            A list containing the names of buffer elements.

        pick_mode:
            The default mode to pick items from the list.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.

        Raises
        ------
        ValueError:
            Cannot modify `buffer_size`. The value is already set.

        ValueError:
            Cannot modify `buffer_names`. The value is already set.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        if buffer_size is not None:
            if self._buffer_size not in (None, buffer_size):
                raise ValueError(
                    'Cannot modify buffer_size. The value is already set.')
            else:
                self._buffer_size = buffer_size

        if buffer_names is not None:
            if self._buffer_names not in (None, buffer_names):
                raise ValueError(
                    'Cannot modify buffer_names. The value is already set.')
            else:
                self._buffer_names = buffer_names

        if self._buffer_size is not None and self._buffer_names is not None:
            self._buffer = {name: [None]*self._buffer_size  # type: ignore
                            for name in self._buffer_names}
        else:
            self._buffer = None

        if pick_mode is not None:
            self._pick_mode = pick_mode

        if clear_buffer is not None:
            self._clear_buffer = clear_buffer

        self.reset()

    def add(self, data: Dict[str, Union[T1, T2]]) -> None:
        '''
        Append a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.

        Notes
        -----
        This implementation of `add` does not check if the buffer is full
        or if the provided names exist in the buffer queues. As a result, this
        situations will result in exceptions by the system.
        '''
        if self._buffer is None:
            raise RuntimeError('Buffer is not set up!')

        self._buffer_index += 1
        for key, v in data.items():
            self._buffer[key][self._buffer_index] = v  # type: ignore

        self._count += 1

    def pick(
            self,
            count: Optional[int] = None,
            mode: Optional[PickModes] = None
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return items from the buffer.

        Arguments
        ---------
        count:
            The number of items to return. If omitted, the number of items in
            the buffer is used. `count` will be ignored if `mode` is 'all'.

        mode:
            How to pick items. If omitted, the default `pick_mode` specified
            during initialization or setup is used.

        Returns
        -------
        :
            A dictionary with buffer names as keys and picked items as values.

        Raises
        ------
        ValueError:
            The number of items requested is greater than the number of items
            in the buffer.

        ValueError:
            `mode` is not one of 'all', 'random', 'recent' and 'old'.
        '''
        _mode = mode.lower() if mode is not None else self._pick_mode
        _count = count if count is not None else self._count

        if _count > self._count:
            raise ValueError('Not enough data in the buffer.')

        if _mode == 'old':
            return self._pick_old(_count)
        elif _mode == 'recent':
            return self._pick_recent(_count)
        elif _mode == 'random':
            return self._pick_random(_count)
        elif _mode == 'all':
            return self._pick_all()
        else:
            raise ValueError(
                'mode should be one of all, old, recent, or random.')

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        if self._clear_buffer is not False:
            self._buffer_index = -1
            self._count = 0

    def _pick_old(
        self, count: int
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return the oldest items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` oldest items
            as values.
        '''
        if self._buffer:
            s = slice(count)
            return {name: tuple(buffer[s])  # type: ignore
                    for name, buffer in self._buffer.items()}
        else:
            raise RuntimeError('Buffer is not set up.')

    def _pick_recent(
        self, count: int
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return the most recent items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` most recent
            items as values.
        '''
        if self._buffer:
            s = slice(self._buffer_index - count + 1, self._buffer_index + 1)
            return {name: tuple(buffer[s])  # type: ignore
                    for name, buffer in self._buffer.items()}
        else:
            raise RuntimeError('Buffer is not set up.')

    def _pick_random(
        self, count: int
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return a random sample of items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` randomly picked
            items as values.
        '''
        if self._buffer:
            index = np.random.choice(
                self._count, count, replace=False)
            return {name: tuple(buffer[i] for i in index)  # type: ignore
                    for name, buffer in self._buffer.items()}
        else:
            raise RuntimeError('Buffer is not set up.')

    def _pick_all(self) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return all items in the buffer.

        Returns
        -------
        :
            A dictionary with buffer names as keys and all items as values.
        '''
        if self._buffer:
            s = slice(self._buffer_index + 1)
            return {name: tuple(buffer[s])  # type: ignore
                    for name, buffer in self._buffer.items()}
        else:
            raise RuntimeError('Buffer is not set up.')

    def dump(self):
        pass
